"""The image configuration and utility manager."""

import math
import warnings
from functools import cached_property
from typing import Any, Literal, TypedDict

import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import Array, Float

from .._internal import error_if_not_positive
from ..constants import (
    interaction_constant_from_kilovolts,
    lorentz_factor_from_kilovolts,
    wavelength_from_kilovolts,
)
from ..jax_util import FloatLike
from ..ndimage import make_coordinate_grid, make_frequency_grid


_get_deprecation_msg = lambda self, prop, func: (
    f"`{self.__class__.__name__}.{prop}` has been deprecated and will be "
    "removed in cryoJAX 0.6.0. Instead, make "
    f"the appropriate call to `{self.__class__.__name__}.{func}`."
)


# Not currently public API
class PrecomputedGrids(eqx.Module, strict=True):
    only_rfft: bool
    only_fourier: bool

    _frequency_grid: Float[Array, "_ _ 2"]
    _coordinate_grid: Float[Array, "_ _ 2"] | None
    _full_frequency_grid: Float[Array, "_ _ 2"] | None

    _padded_coordinate_grid: Float[Array, "_ _ 2"] | None
    _padded_frequency_grid: Float[Array, "_ _ 2"] | None
    _padded_full_frequency_grid: Float[Array, "_ _ 2"] | None

    def __init__(
        self,
        shape: tuple[int, int],
        padded_shape: tuple[int, int] | None = None,
        only_fourier: bool = True,
        only_rfft: bool = True,
    ):
        if only_fourier:
            self._coordinate_grid = None
        else:
            self._coordinate_grid = make_coordinate_grid(shape)
        self._frequency_grid = make_frequency_grid(shape, outputs_rfftfreqs=True)
        if only_rfft:
            self._full_frequency_grid = None
        else:
            self._full_frequency_grid = make_frequency_grid(
                shape, outputs_rfftfreqs=False
            )
        if padded_shape is None or padded_shape == shape:
            self._padded_coordinate_grid = None
            self._padded_frequency_grid = None
            self._padded_full_frequency_grid = None
        else:
            if only_fourier:
                self._padded_coordinate_grid = None
            else:
                self._padded_coordinate_grid = make_coordinate_grid(padded_shape)
            self._padded_frequency_grid = make_frequency_grid(
                padded_shape, outputs_rfftfreqs=True
            )
            if only_rfft:
                self._padded_full_frequency_grid = None
            else:
                self._padded_full_frequency_grid = make_frequency_grid(
                    padded_shape, outputs_rfftfreqs=False
                )
        self.only_fourier = only_fourier
        self.only_rfft = only_rfft

    def get(
        self, *, real_space: bool, full: bool = False, padding: bool = False
    ) -> Array:
        _error_msg = (
            "Internal cryoJAX error when fetching "
            "precomputed grids in the `image_config`. "
            "Please report this issue."
        )
        if padding:
            if real_space:
                _coordinate_grid = (
                    self._coordinate_grid
                    if self._padded_coordinate_grid is None
                    else self._padded_coordinate_grid
                )
                if _coordinate_grid is None:
                    raise Exception(_error_msg)
                return _coordinate_grid
            else:
                if full:
                    _full_frequency_grid = (
                        self._full_frequency_grid
                        if self._padded_full_frequency_grid is None
                        else self._padded_full_frequency_grid
                    )
                    if _full_frequency_grid is None:
                        raise Exception(_error_msg)
                    return _full_frequency_grid
                else:
                    return (
                        self._frequency_grid
                        if self._padded_frequency_grid is None
                        else self._padded_frequency_grid
                    )
        else:
            if real_space:
                if self._coordinate_grid is None:
                    raise Exception(_error_msg)
                return self._coordinate_grid
            else:
                if full:
                    if self._full_frequency_grid is None:
                        raise Exception(_error_msg)
                    return self._full_frequency_grid
                else:
                    return self._frequency_grid


class PadOptions(TypedDict):
    shape: tuple[int, int]


class AbstractImageConfig(eqx.Module, strict=True):
    """Configuration and utilities for an electron microscopy image."""

    shape: eqx.AbstractVar[tuple[int, int]]
    pixel_size: eqx.AbstractVar[Float[Array, ""]]
    voltage_in_kilovolts: eqx.AbstractVar[Float[Array, ""]]

    pad_options: eqx.AbstractVar[dict[str, Any]]

    precompute_mode: eqx.AbstractVar[
        Literal["none", "rfft", "fft", "all", "compile_time_eval"]
    ]
    precomputed_grids: eqx.AbstractVar[PrecomputedGrids | None]

    def __check_init__(self):
        cls = self.__class__.__name__
        if not all(type(s) == int for s in self.padded_shape):
            raise AttributeError(
                f"Found that `{cls}.padded_shape` was not a tuple of Python integers. "
            )
        if not all(type(s) == int for s in self.shape):
            raise AttributeError(
                f"Found that `{cls}.shape` was not a tuple of Python integers. "
            )
        if self.padded_shape[0] < self.shape[0] or self.padded_shape[1] < self.shape[1]:
            raise AttributeError(
                f"Found that `{cls}.padded_shape` is less than `{cls}.shape` in one or "
                " more dimensions."
            )
        if set(self.pad_options.keys()) != {"shape"}:
            raise AttributeError(
                f"Found that `{cls}.pad_options` was not a dictionary with "
                "key 'shape'. "
                f"Instead, had keys {set(self.pad_options.keys())}."
            )

    @property
    def wavelength_in_angstroms(self) -> Float[Array, ""]:
        """The incident electron wavelength corresponding to the beam
        energy `voltage_in_kilovolts`.
        """
        return wavelength_from_kilovolts(error_if_not_positive(self.voltage_in_kilovolts))

    @property
    def lorentz_factor(self) -> Float[Array, ""]:
        """The lorenz factor at the given `voltage_in_kilovolts`."""
        return lorentz_factor_from_kilovolts(
            error_if_not_positive(self.voltage_in_kilovolts)
        )

    @property
    def interaction_constant(self) -> Float[Array, ""]:
        """The electron interaction constant at the given `voltage_in_kilovolts`."""
        return interaction_constant_from_kilovolts(
            error_if_not_positive(self.voltage_in_kilovolts)
        )

    def get_coordinate_grid(self, *, padding: bool = False, physical: bool = True):
        """Return the image coordinate system. See
        [`cryojax.ndimage.make_coordinate_grid`][] for more
        information.

        **Arguments:**

        - `padding`:
            If `True`, return coordinates with shape
            `image_config.padded_shape`. Otherwise, return with
            shape `image_config.shape`.
        - `physical`:
            If `True`, return coordinates in units of angstroms.
            Otherwise, return on the unit box.

        **Returns:**

        The coordinate grid.
        """

        def _get_grid_impl(_self):
            if padding:
                return _self._padded_coordinate_grid
            else:
                return _self._coordinate_grid

        if self.precompute_mode == "compile_time_eval":
            with jax.ensure_compile_time_eval():
                coordinate_grid = _get_grid_impl(self)
        else:
            coordinate_grid = _get_grid_impl(self)

        if physical:
            pixel_size = error_if_not_positive(self.pixel_size)
            coordinate_grid = _safe_multiply_by_constant(coordinate_grid, pixel_size)

        return coordinate_grid

    def get_frequency_grid(
        self, *, padding: bool = False, physical: bool = True, full: bool = False
    ):
        """Return a grid of FFT frequencies. See
        [`cryojax.ndimage.make_frequency_grid`] for more
        information.

        **Arguments:**

        - `padding`:
            If `True`, return frequencies corresponding to shape
            `image_config.padded_shape`. Otherwise, use
            `image_config.shape`.
        - `physical`:
            If `True`, return frequencies in units of inverse
            angstroms. Otherwise, return unitless where Nyquist
            is equal to 0.5.
        - `full`:
            If `True`, return the full plane of frequencies
            for usage with `jax.numpy.fft.fftn`.
            Otherwise, return the half plane for usage with
            `jax.numpy.fft.rfftn`.

        **Returns:**

        The frequency grid.
        """

        def _get_grid_impl(_self):
            if padding:
                if full:
                    return _self._padded_full_frequency_grid
                else:
                    return _self._padded_frequency_grid
            else:
                if full:
                    return _self._full_frequency_grid
                else:
                    return _self._frequency_grid

        if self.precompute_mode == "compile_time_eval":
            with jax.ensure_compile_time_eval():
                frequency_grid = _get_grid_impl(self)
        else:
            frequency_grid = _get_grid_impl(self)

        if physical:
            pixel_size = error_if_not_positive(self.pixel_size)
            frequency_grid = _safe_multiply_by_constant(frequency_grid, 1 / pixel_size)

        return frequency_grid

    @property
    def padded_shape(self):
        return self.pad_options["shape"]

    @property
    def n_pixels(self) -> int:
        """Convenience property for `math.prod(shape)`"""
        return math.prod(self.shape)

    @property
    def y_dim(self) -> int:
        """Convenience property for `shape[0]`"""
        return self.shape[0]

    @property
    def x_dim(self) -> int:
        """Convenience property for `shape[1]`"""
        return self.shape[1]

    @property
    def padded_y_dim(self) -> int:
        """Convenience property for `padded_shape[0]`"""
        return self.padded_shape[0]

    @property
    def padded_x_dim(self) -> int:
        """Convenience property for `padded_shape[1]`"""
        return self.padded_shape[1]

    @property
    def padded_n_pixels(self) -> int:
        """Convenience property for `math.prod(padded_shape)`"""
        return math.prod(self.padded_shape)

    @cached_property
    def _coordinate_grid(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim} 2"]:
        """A spatial coordinate system for the `shape`."""
        if self.precomputed_grids is None or self.precomputed_grids.only_fourier:
            return make_coordinate_grid(self.shape)
        else:
            return self.precomputed_grids.get(real_space=True)

    @cached_property
    def _frequency_grid(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim//2+1} 2"]:
        """A spatial frequency coordinate system for the `shape`,
        with hermitian symmetry.
        """
        if self.precomputed_grids is None:
            return make_frequency_grid(self.shape, outputs_rfftfreqs=True)
        else:
            return self.precomputed_grids.get(real_space=False)

    @cached_property
    def _full_frequency_grid(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim} 2"]:
        """A spatial frequency coordinate system for the `shape`,
        without hermitian symmetry.
        """
        if self.precomputed_grids is None or self.precomputed_grids.only_rfft:
            return make_frequency_grid(shape=self.shape, outputs_rfftfreqs=False)
        else:
            return self.precomputed_grids.get(real_space=False, full=True)

    @cached_property
    def _padded_coordinate_grid(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim} 2"]:
        """A spatial coordinate system for the `padded_shape`."""
        if self.precomputed_grids is None or self.precomputed_grids.only_fourier:
            return make_coordinate_grid(shape=self.padded_shape)
        else:
            return self.precomputed_grids.get(real_space=True, padding=True)

    @cached_property
    def _padded_frequency_grid(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim//2+1} 2"]:
        """A spatial frequency coordinate system for the `padded_shape`,
        with hermitian symmetry.
        """
        if self.precomputed_grids is None:
            return make_frequency_grid(shape=self.padded_shape, outputs_rfftfreqs=True)
        else:
            return self.precomputed_grids.get(real_space=False, padding=True)

    @cached_property
    def _padded_full_frequency_grid(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim} 2"]:
        """A spatial frequency coordinate system for the `padded_shape`,
        without hermitian symmetry.
        """
        if self.precomputed_grids is None or self.precomputed_grids.only_rfft:
            return make_frequency_grid(shape=self.padded_shape, outputs_rfftfreqs=False)
        else:
            return self.precomputed_grids.get(real_space=False, full=True, padding=True)

    @property
    def coordinate_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim} 2"]:
        warnings.warn(
            _get_deprecation_msg(
                self, "coordinate_grid_in_pixels", "get_coordinate_grid"
            ),
            category=FutureWarning,
            stacklevel=2,
        )
        return self._coordinate_grid

    @property
    def coordinate_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim} 2"]:
        warnings.warn(
            _get_deprecation_msg(
                self, "coordinate_grid_in_angstroms", "get_coordinate_grid"
            ),
            category=FutureWarning,
            stacklevel=2,
        )
        return self.get_coordinate_grid()

    @property
    def frequency_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim//2+1} 2"]:
        warnings.warn(
            _get_deprecation_msg(self, "frequency_grid_in_pixels", "get_frequency_grid"),
            category=FutureWarning,
            stacklevel=2,
        )
        return self._frequency_grid

    @property
    def frequency_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim//2+1} 2"]:
        warnings.warn(
            _get_deprecation_msg(
                self, "frequency_grid_in_angstroms", "get_frequency_grid"
            ),
            category=FutureWarning,
            stacklevel=2,
        )
        return self.get_frequency_grid()

    @property
    def full_frequency_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim} 2"]:
        warnings.warn(
            _get_deprecation_msg(
                self, "full_frequency_grid_in_pixels", "get_frequency_grid"
            ),
            category=FutureWarning,
            stacklevel=2,
        )
        return self._full_frequency_grid

    @property
    def full_frequency_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim} 2"]:
        warnings.warn(
            _get_deprecation_msg(
                self, "full_frequency_grid_in_angstroms", "get_frequency_grid"
            ),
            category=FutureWarning,
            stacklevel=2,
        )
        return self.get_frequency_grid(full=True)

    @property
    def padded_coordinate_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim} 2"]:
        warnings.warn(
            _get_deprecation_msg(
                self, "padded_coordinate_grid_in_pixels", "get_coordinate_grid"
            ),
            category=FutureWarning,
            stacklevel=2,
        )
        return self._padded_coordinate_grid

    @property
    def padded_coordinate_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim} 2"]:
        warnings.warn(
            _get_deprecation_msg(
                self, "coordinate_grid_in_angstroms", "get_coordinate_grid"
            ),
            category=FutureWarning,
            stacklevel=2,
        )
        return self.get_coordinate_grid(padding=True)

    @property
    def padded_frequency_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim//2+1} 2"]:
        warnings.warn(
            _get_deprecation_msg(
                self, "padded_frequency_grid_in_pixels", "get_frequency_grid"
            ),
            category=FutureWarning,
            stacklevel=2,
        )
        return self._padded_frequency_grid

    @property
    def padded_frequency_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim//2+1} 2"]:
        warnings.warn(
            _get_deprecation_msg(
                self, "padded_frequency_grid_in_angstroms", "get_frequency_grid"
            ),
            category=FutureWarning,
            stacklevel=2,
        )
        return self.get_frequency_grid(padding=True)

    @property
    def padded_full_frequency_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim} 2"]:
        warnings.warn(
            _get_deprecation_msg(
                self, "padded_full_frequency_grid_in_pixels", "get_frequency_grid"
            ),
            category=FutureWarning,
            stacklevel=2,
        )
        return self._padded_full_frequency_grid

    @property
    def padded_full_frequency_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim} 2"]:
        warnings.warn(
            _get_deprecation_msg(
                self, "padded_full_frequency_grid_in_angstroms", "get_frequency_grid"
            ),
            category=FutureWarning,
            stacklevel=2,
        )
        return self.get_frequency_grid(full=True, padding=True)


class BasicImageConfig(AbstractImageConfig, strict=True):
    """Configuration and utilities for a basic electron microscopy
    image.
    """

    shape: tuple[int, int]
    pixel_size: Float[Array, ""]
    voltage_in_kilovolts: Float[Array, ""]

    pad_options: PadOptions

    precompute_mode: Literal["none", "rfft", "fft", "all", "compile_time_eval"]
    precomputed_grids: PrecomputedGrids | None

    def __init__(
        self,
        shape: tuple[int, int],
        pixel_size: FloatLike,
        voltage_in_kilovolts: FloatLike,
        *,
        precompute_mode: Literal[
            "none", "rfft", "fft", "all", "compile_time_eval"
        ] = "none",
        pad_options: dict[str, Any] = {},
    ):
        """**Arguments:**

        - `shape`:
            Shape of the imaging plane in pixels.
        - `pixel_size`:
            The pixel size of the image in angstroms.
        - `voltage_in_kilovolts`:
            The incident energy of the electron beam.
        - `pad_options`:
            Options that control image padding.
            - 'shape':
                The shape of the image after padding. By default, equal
                to `shape`.
        - `precompute_mode`:
            How to pre-compute coordinate and frequency grids stored in
            the `image_config`. Options are
            - 'none':
                Compute grids at runtime and cache the result.
            - 'rfft':
                Only precompute frequencies
                in the Fourier domain for a real-valued function (i.e.
                for use with `jax.numpy.fft.rfftn`). This is the best option
                for most use cases.
            - 'fft':
                Precompute frequencies
                in the Fourier domain for both real and complex-valued functions
                (i.e. for use with `jax.numpy.fft.rfftn` and `jax.numpy.fft.fftn`).
                Relevant for use with Ewald sphere extraction.
            - 'all':
                Precompute all grids in both real-space and frequency-space.
            - 'compile_time_eval':
                Evaluate grids as needed at compile time using
                `jax.ensure_compile_time_eval`.
        """
        # Set parameters
        self.pixel_size = jnp.asarray(pixel_size, dtype=float)
        self.voltage_in_kilovolts = jnp.asarray(voltage_in_kilovolts, dtype=float)
        # Set shape
        self.shape = shape
        # Set pad options
        self.pad_options = _dict_to_pad_options(pad_options, shape)
        # Finally, grid precompute
        if precompute_mode == "rfft":
            self.precomputed_grids = PrecomputedGrids(
                self.shape, self.padded_shape, only_rfft=True
            )
        elif precompute_mode == "fft":
            self.precomputed_grids = PrecomputedGrids(
                self.shape, self.padded_shape, only_rfft=False
            )
        elif precompute_mode == "all":
            self.precomputed_grids = PrecomputedGrids(
                self.shape, self.padded_shape, only_fourier=False, only_rfft=False
            )
        else:
            self.precomputed_grids = None
        self.precompute_mode = precompute_mode


class DoseImageConfig(AbstractImageConfig, strict=True):
    """Configuration and utilities for an electron microscopy image,
    including the electron dose."""

    shape: tuple[int, int]
    pixel_size: Float[Array, ""]
    voltage_in_kilovolts: Float[Array, ""]
    electron_dose: Float[Array, ""]

    pad_options: PadOptions

    precompute_mode: Literal["none", "rfft", "fft", "all", "compile_time_eval"]
    precomputed_grids: PrecomputedGrids | None

    def __init__(
        self,
        shape: tuple[int, int],
        pixel_size: FloatLike,
        voltage_in_kilovolts: FloatLike,
        electron_dose: FloatLike,
        *,
        precompute_mode: Literal[
            "none", "rfft", "fft", "all", "compile_time_eval"
        ] = "none",
        pad_options: dict[str, Any] = {},
    ):
        """**Arguments:**

        - `shape`:
            Shape of the imaging plane in pixels.
        - `pixel_size`:
            The pixel size of the image in angstroms.
        - `voltage_in_kilovolts`:
            The incident energy of the electron beam.
        - `electron_dose`:
            The integrated dose rate of the electron beam in
            $e^-/A^2$
        - `pad_options`:
            Options that control image padding. This is a dictionary
            with keys:
            - 'shape':
                The shape of the image after padding. By default, equal
                to `shape`.
        - `precompute_mode`:
            How to pre-compute coordinate and frequency grids stored in
            the `image_config`. Options are
            - 'none':
                Compute grids at runtime and cache the result.
            - 'rfft':
                Only precompute frequencies
                in the Fourier domain for a real-valued function (i.e.
                for use with `jax.numpy.fft.rfftn`). This is the best option
                for most use cases.
            - 'fft':
                Precompute frequencies
                in the Fourier domain for both real and complex-valued functions
                (i.e. for use with `jax.numpy.fft.rfftn` and `jax.numpy.fft.fftn`).
                Relevant for use with Ewald sphere extraction.
            - 'all':
                Precompute all grids in both real-space and frequency-space.
            - 'compile_time_eval':
                Evaluate grids as needed at compile time using
                `jax.ensure_compile_time_eval`.
        """
        # Set parameters
        self.pixel_size = jnp.asarray(pixel_size, dtype=float)
        self.voltage_in_kilovolts = jnp.asarray(voltage_in_kilovolts, dtype=float)
        self.electron_dose = jnp.asarray(electron_dose, dtype=float)
        # Set shape
        self.shape = shape
        # Set pad options
        self.pad_options = _dict_to_pad_options(pad_options, shape)
        # Finally, grid precompute
        if precompute_mode == "rfft":
            self.precomputed_grids = PrecomputedGrids(
                self.shape, self.padded_shape, only_rfft=True
            )
        elif precompute_mode == "fft":
            self.precomputed_grids = PrecomputedGrids(
                self.shape, self.padded_shape, only_rfft=False
            )
        elif precompute_mode == "all":
            self.precomputed_grids = PrecomputedGrids(
                self.shape, self.padded_shape, only_fourier=False, only_rfft=False
            )
        else:
            self.precomputed_grids = None
        self.precompute_mode = precompute_mode

    @property
    def electrons_per_pixel(self) -> Float[Array, ""]:
        """The `electron_dose` in a given pixel area."""
        return error_if_not_positive(self.electron_dose) * self.pixel_size**2


def _safe_multiply_by_constant(
    grid: Float[Array, "y_dim x_dim 2"], constant: Float[Array, ""]
) -> Float[Array, "y_dim x_dim 2"]:
    """Multiplies a coordinate grid by a constant in a
    safe way for gradient computation.
    """
    return jnp.where(grid != 0.0, constant * grid, 0.0)


def _dict_to_pad_options(d: dict[str, Any], default_shape: tuple[int, int]) -> PadOptions:
    shape = d["shape"] if "shape" in d else default_shape

    return PadOptions(shape=shape)



import warnings

import cryojax.ndimage as im
import cryojax.simulator as cxs
import equinox as eqx
import numpy as np
import pytest
from cryojax.atom_util import split_atoms_by_element
from cryojax.constants import PengScatteringFactorParameters
from cryojax.io import read_atoms_from_pdb
from jaxtyping import Array


try:
    import jax_finufft as jnufft

    JAX_FINUFFT_IMPORT_ERROR = None
except ModuleNotFoundError as err:
    jnufft = None
    JAX_FINUFFT_IMPORT_ERROR = err


@pytest.fixture
def pdb_info(sample_pdb_path):
    return read_atoms_from_pdb(sample_pdb_path, center=True, loads_properties=True)


@pytest.mark.parametrize("shape", ((64, 64), (63, 63), (63, 64), (64, 63)))
def test_gmm_integrator_shape(sample_pdb_path, shape):
    atom_volume = cxs.load_tabulated_volume(
        sample_pdb_path,
        output_type=cxs.GaussianMixtureVolume,
        include_b_factors=True,
        selection_string="not element H",
    )
    pixel_size = 0.5

    integrator = cxs.GaussianMixtureProjection(shape=(2 * shape[0], 2 * shape[1]))
    # # ... and the configuration of the imaging instrument
    image_config = cxs.BasicImageConfig(
        shape=shape,
        pixel_size=pixel_size,
        voltage_in_kilovolts=300.0,
    )
    # ... compute the integrated volume
    fourier_integrated_volume = integrator.integrate(
        atom_volume, image_config, outputs_real_space=False
    )

    assert fourier_integrated_volume.shape == (shape[0], shape[1] // 2 + 1)


def test_fft_atom_bad_instantiation():
    with pytest.raises(ValueError):
        _ = cxs.IndependentAtomVolume(
            positions=np.zeros((10, 3)),
            scattering_factors=(im.FourierGaussian(),),
        )
    # with pytest.raises(ValueError):
    #     _ = cxs.FFTAtomProjection(upsample_factor=2)


@pytest.mark.parametrize(
    "pixel_size, shape",
    ((1.0, (32, 32)), (1.0, (32, 31)), (1.0, (31, 32)), (1.0, (31, 31))),
)
def test_fft_atom_projection_exact(pdb_info, pixel_size, shape):
    if jnufft is not None:
        atom_positions, _, _ = pdb_info
        pixel_size, shape = 0.5, (64, 64)
        pad_options = dict(shape=(128, 128))
        image_config = cxs.BasicImageConfig(
            shape, pixel_size, voltage_in_kilovolts=300.0, pad_options=pad_options
        )
        amplitude, b_factor = 1.0, 100.0
        gaussian_volume, gaussian_integrator = (
            cxs.GaussianMixtureVolume(
                atom_positions,
                amplitudes=amplitude,
                variances=b_factor / (8 * np.pi**2),
            ),
            cxs.GaussianMixtureProjection(sampling_mode="point"),
        )
        atom_volume, fft_integrator = (
            cxs.IndependentAtomVolume(
                positions=atom_positions,
                scattering_factors=im.FourierGaussian(
                    amplitude=amplitude, b_factor=b_factor
                ),
            ),
            cxs.FFTAtomProjection(sampling_mode="point", eps=1e-16),
        )
        proj_by_gaussians = compute_projection(
            gaussian_volume, gaussian_integrator, image_config
        )
        proj_by_fft = compute_projection(atom_volume, fft_integrator, image_config)
        np.testing.assert_allclose(proj_by_gaussians, proj_by_fft, atol=1e-8)
    else:
        warnings.warn(
            "Could not test projection method `FFTAtomProjection`, "
            "most likely because `jax_finufft` is not installed. "
            f"Error traceback is:\n{JAX_FINUFFT_IMPORT_ERROR}"
        )


@pytest.mark.parametrize(
    "width, pixel_size, shape",
    ((5.0, 0.5, (64, 64)), (1.0, 0.5, (64, 64)), (2.0, 1.0, (32, 32))),
)
def test_fft_atom_projection_antialias(pdb_info, width, pixel_size, shape):
    if jnufft is not None:
        atom_positions, _, _ = pdb_info
        gaussian_volume = cxs.GaussianMixtureVolume(
            atom_positions,
            amplitudes=1.0,
            variances=width**2,
        )
        atom_volume = cxs.IndependentAtomVolume(
            positions=atom_positions,
            scattering_factors=im.FourierGaussian(
                amplitude=1.0, b_factor=width**2 * (8 * np.pi**2)
            ),
        )
        gaussian_integrator = cxs.GaussianMixtureProjection(sampling_mode="average")
        fft_integrator = cxs.FFTAtomProjection(eps=1e-16)
        pad_options = dict(shape=(2 * shape[0], 2 * shape[1]))
        image_config = cxs.BasicImageConfig(
            shape, pixel_size, voltage_in_kilovolts=300.0, pad_options=pad_options
        )
        proj_by_gaussians = compute_projection(
            gaussian_volume, gaussian_integrator, image_config
        )
        proj_by_fft = compute_projection(atom_volume, fft_integrator, image_config)

        np.testing.assert_allclose(proj_by_gaussians, proj_by_fft, atol=1e-8)


@pytest.mark.parametrize(
    "pixel_size, shape, upsample_factor",
    (
        (0.25, (128, 128), 4),
        (0.25, (127, 127), 4),
        (0.25, (128, 128), 5),
        (0.25, (127, 127), 5),
        (0.25, (127, 128), 5),
        (0.25, (128, 127), 5),
    ),
)
def test_fft_atom_projection_peng(pdb_info, pixel_size, shape, upsample_factor):
    if jnufft is not None:
        atom_positions, atom_ids, _ = pdb_info
        positions_by_id, unique_atom_ids = split_atoms_by_element(
            atom_ids, atom_positions
        )
        peng_parameters, peng_parameters_by_id = (
            PengScatteringFactorParameters(atom_ids),
            PengScatteringFactorParameters(unique_atom_ids),
        )
        gaussian_volume = cxs.GaussianMixtureVolume.from_tabulated_parameters(
            atom_positions,
            peng_parameters,
        )
        atom_volume = cxs.IndependentAtomVolume.from_tabulated_parameters(
            positions_by_id,
            peng_parameters_by_id,
        )
        image_config = cxs.BasicImageConfig(shape, pixel_size, voltage_in_kilovolts=300.0)
        # Check to make sure the implementations are identical, up to the
        # nufft (don't include anti-aliasing)
        gaussian_integrator = cxs.GaussianMixtureProjection(sampling_mode="average")
        fft_integrator = cxs.FFTAtomProjection(
            sampling_mode="average", upsample_factor=upsample_factor, eps=1e-16
        )
        proj_by_gaussians = compute_projection(
            gaussian_volume, gaussian_integrator, image_config
        )
        proj_by_fft = compute_projection(atom_volume, fft_integrator, image_config)
        np.testing.assert_allclose(proj_by_gaussians, proj_by_fft, atol=5e-3)


@pytest.mark.parametrize(
    "pixel_size, shape",
    (
        (1.0, (32, 32)),
        (1.0, (31, 31)),
        (1.0, (31, 32)),
        (1.0, (32, 31)),
    ),
)
def test_analytic_vs_voxels_nopose(pdb_info, pixel_size, shape):
    """
    Test that computing a projection analytically with gaussians
    agrees with numerical results for voxel-based volumes.
    """
    # Unpack PDB info
    atom_positions, atom_types, atom_properties = pdb_info
    # Objects for imaging
    image_config = cxs.BasicImageConfig(
        shape,
        pixel_size,
        voltage_in_kilovolts=300.0,
    )
    # Real vs fourier volumes
    dim = max(*shape)  # Make sure to use `padded_shape` here

    peng_parameters = PengScatteringFactorParameters(atom_types)
    base_volume = cxs.GaussianMixtureVolume.from_tabulated_parameters(
        atom_positions,
        peng_parameters,
        extra_b_factors=atom_properties["b_factors"],
    )
    base_method = cxs.GaussianMixtureProjection(sampling_mode="average")
    render_volume_fn = cxs.GaussianMixtureRenderFn((dim, dim, dim), pixel_size)
    real_voxel_grid = render_volume_fn(base_volume)
    other_volumes = [
        cxs.FourierVoxelGridVolume.from_real_voxel_grid(real_voxel_grid),
        make_spline(real_voxel_grid),
    ]
    other_projection_methods = [
        cxs.FourierSliceExtraction(),
        cxs.FourierSliceExtraction(),
    ]
    if jnufft is not None:
        other_projection_methods.append(cxs.RealVoxelProjection(eps=1e-16))  # type: ignore
        other_volumes.append(
            cxs.RealVoxelGridVolume.from_real_voxel_grid(real_voxel_grid)
        )
    else:
        warnings.warn(
            "Could not test projection method `RealVoxelProjection`, "
            "most likely because `jax_finufft` is not installed. "
            f"Error traceback is:\n{JAX_FINUFFT_IMPORT_ERROR}"
        )

    projection_by_gaussian_integration = compute_projection(
        base_volume, base_method, image_config
    )
    for volume, projection_method in zip(other_volumes, other_projection_methods):
        projection_by_other_method = compute_projection(
            volume, projection_method, image_config
        )
        np.testing.assert_allclose(
            projection_by_gaussian_integration, projection_by_other_method, atol=1e-12
        )


# @pytest.mark.parametrize(
#     "pixel_size, shape, euler_pose_params",
#     (
#         (1.0, (32, 32), (2.5, -5.0, 0.0, 0.0, 0.0)),
#         (1.0, (32, 32), (0.0, 0.0, 10.0, -30.0, 60.0)),
#         (1.0, (32, 32), (2.5, -5.0, 10.0, -30.0, 60.0)),
#     ),
# )
# def test_projection_methods_with_pose(
#     sample_pdb_path, pixel_size, shape, euler_pose_params
# ):
#     """Test that computing a projection across different
#     methods agrees. This tests pose convention and accuracy
#     for real vs fourier, atoms vs voxels, etc.
#     """
#     # Objects for imaging
#     instrument_config = cxs.BasicImageConfig(
#         shape,
#         pixel_size,
#         voltage_in_kilovolts=300.0,
#     )
#     euler_pose = cxs.EulerAnglePose(*euler_pose_params)
#     # Real vs fourier potentials
#     dim = max(*shape)
#     atom_positions, atom_types, b_factors = read_atoms_from_pdb(
#         sample_pdb_path, center=True, loads_b_factors=True
#     )
#     scattering_factor_parameters = get_tabulated_scattering_factor_parameters(
#         atom_types, read_peng_element_scattering_factor_parameter_table()
#     )
#     base_potential = cxs.PengAtomicPotential(
#         atom_positions,
#         scattering_factor_a=scattering_factor_parameters["a"],
#         scattering_factor_b=scattering_factor_parameters["b"],
#         b_factors=b_factors,
#     )
#     base_method = cxs.GaussianMixtureProjection(sampling_mode="average")

#     real_voxel_grid = base_potential.as_real_voxel_grid((dim, dim, dim), pixel_size)
#     other_potentials = [
#         cxs.FourierVoxelGridPotential.from_real_voxel_grid(real_voxel_grid, pixel_size),
#         make_spline_potential(real_voxel_grid, pixel_size),
#         cxs.GaussianMixtureAtomicPotential(
#             atom_positions,
#             scattering_factor_parameters["a"],
#             (scattering_factor_parameters["b"] + b_factors[:, None]) / (8 * jnp.pi**2),
#         ),
#     ]
#     #     cxs.RealVoxelGridPotential.from_real_voxel_grid(real_voxel_grid, pixel_size),
#     #     cxs.RealVoxelCloudPotential.from_real_voxel_grid(real_voxel_grid, pixel_size),
#     # ]
#     other_projection_methods = [
#         cxs.FourierSliceExtraction(),
#         cxs.FourierSliceExtraction(),
#         base_method,
#     ]
#     #     cxs.NufftProjection(),
#     #     cxs.NufftProjection(),
#     # ]

#     projection_by_gaussian_integration = compute_projection_at_pose(
#         base_potential, base_method, euler_pose, instrument_config
#     )
#     for idx, (potential, projection_method) in enumerate(
#         zip(other_potentials, other_projection_methods)
#     ):
#         if isinstance(projection_method, cxs.NufftProjection):
#             try:
#                 projection_by_other_method = compute_projection_at_pose(
#                     potential, projection_method, euler_pose, instrument_config
#                 )
#             except Exception as err:
#                 warnings.warn(
#                     "Could not test projection method `NufftProjection` "
#                     "This is most likely because `jax_finufft` is not installed. "
#                     f"Error traceback is:\n{err}"
#                 )
#                 continue
#         else:
#             projection_by_other_method = compute_projection_at_pose(
#                 potential, projection_method, euler_pose, instrument_config
#             )
#         np.testing.assert_allclose(
#             np.sum(
#                 (projection_by_gaussian_integration - projection_by_other_method) ** 2
#             ),
#             0.0,
#             atol=1e-8,
#         )


# def plot_images(proj1, proj2):
#     from matplotlib import pyplot as plt
#     from mpl_toolkits.axes_grid1 import make_axes_locatable

#     vmin, vmax = min(proj1.min(), proj2.min()), max(proj1.max(), proj2.max())
#     fig, axes = plt.subplots(figsize=(15, 5), ncols=3)
#     im1 = axes[0].imshow(proj1, vmin=vmin, vmax=vmax, cmap="gray")
#     im2 = axes[1].imshow(proj2, vmin=vmin, vmax=vmax, cmap="gray")
#     for im, ax in zip([im1, im2], axes):
#         divider = make_axes_locatable(ax)
#         cax = divider.append_axes("right", size="5%", pad=0.05)
#         fig.colorbar(im, cax=cax)
#     im3 = axes[2].imshow(np.abs(proj2 - proj1), cmap="gray")
#     divider = make_axes_locatable(axes[2])
#     cax = divider.append_axes("right", size="5%", pad=0.05)
#     fig.colorbar(im3, cax=cax)
#     plt.show()


@eqx.filter_jit
def compute_projection(
    volume: cxs.AbstractVolumeRepresentation,
    integrator: cxs.AbstractVolumeIntegrator,
    image_config: cxs.BasicImageConfig,
) -> Array:
    fourier_projection = integrator.integrate(
        volume, image_config, outputs_real_space=False
    )
    return im.crop_to_shape(
        im.irfftn(
            fourier_projection,
            s=image_config.padded_shape,
        ),
        image_config.shape,
    )


@eqx.filter_jit
def compute_projection_at_pose(
    volume: cxs.AbstractVolumeRepresentation,
    integrator: cxs.AbstractVolumeIntegrator,
    pose: cxs.AbstractPose,
    image_config: cxs.BasicImageConfig,
) -> Array:
    rotated_volume = volume.rotate_to_pose(pose)
    fourier_projection = integrator.integrate(
        rotated_volume, image_config, outputs_real_space=False
    )
    translation_operator = pose.compute_translation_operator(
        image_config.get_frequency_grid(padding=True, physical=True)
    )
    return im.crop_to_shape(
        im.irfftn(
            pose.translate_image(
                fourier_projection,
                translation_operator,
                image_config.padded_shape,
            ),
            s=image_config.padded_shape,
        ),
        image_config.shape,
    )


@eqx.filter_jit
def make_spline(real_voxel_grid):
    return cxs.FourierVoxelSplineVolume.from_real_voxel_grid(
        real_voxel_grid,
    )



from .base_noise_model import (
    AbstractEmpiricalNoiseModel as AbstractEmpiricalNoiseModel,
    AbstractLikelihoodNoiseModel as AbstractLikelihoodNoiseModel,
    AbstractNoiseModel as AbstractNoiseModel,
)
from .gaussian_noise_model import (
    AbstractGaussianNoiseModel as AbstractGaussianNoiseModel,
    GaussianColoredNoiseModel as GaussianColoredNoiseModel,
    GaussianWhiteNoiseModel as GaussianWhiteNoiseModel,
)



import importlib
import re

import cryojax as cx
import cryojax.simulator as cxs
import numpy as np
import pytest
from packaging.version import parse as parse_version


def test_future_deprecated(sample_pdb_path):
    match = re.match(r"(\d+\.\d+(?:\.\d+)?)", cx.__version__)
    assert match, f"Could not parse current cryojax version {cx.__version__!r}"
    current_version = parse_version(match.group(1))

    def should_be_removed(_record):
        msg = str(_record[0].message)
        match = re.search(r"\b(\d+\.\d+(?:\.\d+)?)\b", msg)
        assert match, f"Could not parse removal version from warning message: {msg}"
        removal_version = parse_version(match.group(1))
        return current_version >= removal_version

    ic = cxs.BasicImageConfig((10, 10), 1.5, 300.0, precompute_mode="all")

    with pytest.warns(FutureWarning) as record:
        np.testing.assert_allclose(
            ic.coordinate_grid_in_angstroms, ic.get_coordinate_grid()
        )
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        assert ic.coordinate_grid_in_pixels is ic.get_coordinate_grid(physical=False)
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        assert ic.padded_coordinate_grid_in_pixels is ic.get_coordinate_grid(
            padding=True, physical=False
        )
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        np.testing.assert_allclose(
            ic.padded_coordinate_grid_in_angstroms,
            ic.get_coordinate_grid(
                padding=True,
            ),
        )
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        np.testing.assert_allclose(
            ic.frequency_grid_in_angstroms, ic.get_frequency_grid()
        )
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        assert ic.frequency_grid_in_pixels is ic.get_frequency_grid(physical=False)
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        np.testing.assert_allclose(
            ic.padded_frequency_grid_in_angstroms, ic.get_frequency_grid(padding=True)
        )
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        assert ic.padded_frequency_grid_in_pixels is ic.get_frequency_grid(
            padding=True, physical=False
        )
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        assert ic.full_frequency_grid_in_pixels is ic.get_frequency_grid(
            physical=False, full=True
        )
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        np.testing.assert_allclose(
            ic.full_frequency_grid_in_angstroms,
            ic.get_frequency_grid(physical=True, full=True),
        )
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        assert ic.padded_full_frequency_grid_in_pixels is ic.get_frequency_grid(
            padding=True, physical=False, full=True
        )
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        np.testing.assert_allclose(
            ic.padded_full_frequency_grid_in_angstroms,
            ic.get_frequency_grid(padding=True, physical=True, full=True),
        )
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        obj = cxs.AberratedAstigmaticCTF
        assert obj is cxs.AstigmaticCTF
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        obj = cxs.CTF
        assert obj is cxs.AstigmaticCTF
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        obj = cxs.CorrelatedGaussianNoiseModel
        assert obj is cxs.GaussianColoredNoiseModel
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        obj = cxs.UncorrelatedGaussianNoiseModel
        assert obj is cxs.GaussianWhiteNoiseModel
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        obj = cxs.NufftProjection
        assert obj is cxs.RealVoxelProjection
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        obj = cxs.PengScatteringFactorParameters
        assert obj is cx.constants.PengScatteringFactorParameters
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        obj = cxs.PengAtomicVolume
        assert obj is cxs.GaussianMixtureVolume
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        atom_pos, _, _ = cx.io.read_atoms_from_pdb(  # type: ignore
            sample_pdb_path,
            loads_b_factors=True,
        )
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        volume = cxs.GaussianMixtureVolume(atom_pos, amplitudes=1.0, variances=1.0)
        _ = volume.to_real_voxel_grid((32, 32, 32), 2.0)
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        _ = cxs.GaussianMixtureProjection(use_error_functions=True)  # type: ignore
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        func = cx.ndimage.downsample_with_fourier_cropping
        assert func is cx.ndimage.fourier_crop_downsample
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        func = cx.ndimage.downsample_to_shape_with_fourier_cropping
        assert func is cx.ndimage.fourier_crop_to_shape
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        func = cx.ndimage.normalize_image
        assert func is cx.ndimage.standardize_image
        assert not should_be_removed(record)

    # ndimage submodules
    with pytest.warns(FutureWarning) as record:
        from cryojax.coordinates import make_frequency_grid as make_1
        from cryojax.ndimage import make_frequency_grid as make_2

        assert make_1 is make_2
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        from cryojax.ndimage import AbstractFourierOperator, operators as op

        assert AbstractFourierOperator is op.AbstractFourierOperator
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        from cryojax.ndimage import AbstractFilter, transforms as tf

        assert AbstractFilter is tf.AbstractFilter
        assert not should_be_removed(record)

    with pytest.warns(FutureWarning) as record:
        import jax.numpy as jnp
        from cryojax.jax_util import MinimumSearchMethod, run_grid_search

        def fn(tree_grid_point, _):
            x, y = tree_grid_point
            return (x - 1.0) ** 2 + (y - 2.0) ** 2

        tree_grid = (jnp.arange(-5.0, 5.0, 1), jnp.arange(-5.0, 5.0, 0.1))
        _ = run_grid_search(
            fn, method=MinimumSearchMethod(), tree_grid=tree_grid, args=None
        )


def test_deprecated():
    DEPRECATED = [
        "cryojax.simulator.DiscreteStructuralEnsemble",
    ]

    # Deprecated features
    for path in DEPRECATED:
        mod_path, _, attr = path.rpartition(".")
        module = importlib.import_module(mod_path)
        with pytest.raises(ImportError):
            _ = getattr(module, attr)



from typing_extensions import override

from jaxtyping import Array, Complex, PRNGKeyArray

from .._image_config import AbstractImageConfig
from .._multislice import AbstractMultisliceIntegrator
from .._scattering_theory import AbstractWaveScatteringTheory
from .._transfer_theory import WaveTransferTheory
from .._volume import AbstractVolumeRepresentation


class MultisliceScatteringTheory(AbstractWaveScatteringTheory, strict=True):
    """A scattering theory using the multislice method."""

    volume_integrator: AbstractMultisliceIntegrator
    transfer_theory: WaveTransferTheory

    def __init__(
        self,
        volume_integrator: AbstractMultisliceIntegrator,
        transfer_theory: WaveTransferTheory,
    ):
        """**Arguments:**

        - `volume_integrator`: The multislice method.
        - `transfer_theory`: The wave transfer theory.
        """
        self.volume_integrator = volume_integrator
        self.transfer_theory = transfer_theory

    @override
    def compute_exit_wave(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        rng_key: PRNGKeyArray | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]:
        del rng_key
        # Compute the wavefunction in the exit plane
        wavefunction = self.volume_integrator.integrate(
            volume_representation, image_config
        )

        return wavefunction



"""
Utilities for creating equinox filter_specs.
"""

from collections.abc import Callable, Sequence
from typing import Any

import equinox as eqx
import jax.tree_util as jtu
from jaxtyping import PyTree


def make_filter_spec(
    pytree: PyTree,
    where: Callable[[PyTree], Any | Sequence[Any]],
    *,
    inverse: bool = False,
    is_leaf: Callable[[Any], bool] | None = None,
) -> PyTree[bool]:
    """A lightweight wrapper around `equinox` for creating a "filter specification".

    A filter specification, or `filter_spec`, is a pytree whose
    leaves are either `True` or `False`. These are commonly used with
    `equinox` [filtering](https://docs.kidger.site/equinox/all-of-equinox/#2-filtering).

    In `cryojax`, it is a common pattern to need to finely specify which
    leaves we would like to take JAX transformations with respect to. This is done with a
    pointer to individual leaves, which is referred to as a `where` function. See
    [`here`](https://docs.kidger.site/equinox/examples/frozen_layer/#freezing-parameters)
    in the `equinox` documentation for an example.

    **Returns:**

    The filter specification. This is a pytree of the same structure as `pytree` with
    `True` where the `where` function points to, and `False` where it does not
    (or the opposite, if `inverse = True`).
    """
    if not inverse:
        false_pytree = jtu.tree_map(lambda _: False, pytree, is_leaf=is_leaf)
        return eqx.tree_at(
            where, false_pytree, replace_fn=lambda _: True, is_leaf=is_leaf
        )
    else:
        true_pytree = jtu.tree_map(lambda _: True, pytree)
        return eqx.tree_at(
            where, true_pytree, replace_fn=lambda _: False, is_leaf=is_leaf
        )



"""
Abstraction of electron detectors in a cryo-EM image.
"""

from abc import abstractmethod
from typing_extensions import override

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
import numpy as np
from equinox import Module
from jaxtyping import Array, Complex, Float, PRNGKeyArray

from ..ndimage import irfftn, rfftn
from ._image_config import DoseImageConfig


class AbstractDQE(eqx.Module, strict=True):
    r"""Base class for a detector DQE."""

    @abstractmethod
    def __call__(self, image_config: DoseImageConfig) -> Float[Array, "_ _"]:
        raise NotImplementedError


class NullDQE(AbstractDQE, strict=True):
    r"""A DQE that is perfect across all spatial frequencies."""

    @override
    def __call__(
        self, image_config: DoseImageConfig
    ) -> Float[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        """**Arguments:**

        - `image_config`: The image config.
        """
        return jnp.full(
            (image_config.padded_y_dim, image_config.padded_x_dim // 2 + 1), 1.0
        )


class AbstractDetector(Module, strict=True):
    """Base class for an electron detector."""

    dqe: AbstractDQE

    def __init__(self, dqe: AbstractDQE = NullDQE()):
        self.dqe = dqe

    @abstractmethod
    def sample_readout_from_expected_events(
        self, key: PRNGKeyArray, expected_electron_events: Float[Array, "y_dim x_dim"]
    ) -> Float[Array, "y_dim x_dim"]:
        """Sample a realization from the detector noise model."""
        raise NotImplementedError

    def compute_expected_electron_events(
        self,
        fourier_intensity: Complex[
            Array,
            "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}",
        ],
        image_config: DoseImageConfig,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        """Compute the expected electron events from the detector."""
        fourier_expected_electron_events = (
            self._compute_expected_events_or_detector_readout(
                fourier_intensity, image_config, key=None
            )
        )

        return fourier_expected_electron_events

    def compute_detector_readout(
        self,
        key: PRNGKeyArray,
        fourier_intensity: Complex[
            Array,
            "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}",
        ],
        image_config: DoseImageConfig,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        """Measure the readout from the detector."""
        fourier_detector_readout = self._compute_expected_events_or_detector_readout(
            fourier_intensity, image_config, key
        )

        return fourier_detector_readout

    def _compute_expected_events_or_detector_readout(
        self,
        fourier_intensity: Complex[
            Array,
            "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}",
        ],
        image_config: DoseImageConfig,
        key: PRNGKeyArray | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        """Pass the image through the detector model."""
        # The total number of electrons over the entire image
        n_pixels = np.prod(image_config.padded_shape)
        electrons_per_image = n_pixels * image_config.electrons_per_pixel
        # Normalize the squared wavefunction to a set of probabilities
        fourier_intensity /= fourier_intensity[0, 0]
        # Compute the noiseless signal by applying the DQE to the squared wavefunction
        fourier_signal = fourier_intensity * jnp.sqrt(self.dqe(image_config))
        # Apply the integrated dose rate
        fourier_expected_electron_events = electrons_per_image * fourier_signal
        if key is None:
            # If there is no key given, return
            return fourier_expected_electron_events
        else:
            # ... otherwise, go to real space, sample, go back to fourier,
            # and return.
            expected_electron_events = irfftn(
                fourier_expected_electron_events, s=image_config.padded_shape
            )
            return rfftn(
                self.sample_readout_from_expected_events(key, expected_electron_events)
            )


class GaussianDetector(AbstractDetector, strict=True):
    """A detector with a gaussian noise model. This is the gaussian limit
    of `PoissonDetector`.
    """

    @override
    def sample_readout_from_expected_events(
        self, key: PRNGKeyArray, expected_electron_events: Float[Array, "y_dim x_dim"]
    ) -> Float[Array, "y_dim x_dim"]:
        return expected_electron_events + jnp.sqrt(expected_electron_events) * jr.normal(
            key, expected_electron_events.shape
        )


class PoissonDetector(AbstractDetector, strict=True):
    """A detector with a poisson noise model."""

    @override
    def sample_readout_from_expected_events(
        self, key: PRNGKeyArray, expected_electron_events: Float[Array, "y_dim x_dim"]
    ) -> Float[Array, "y_dim x_dim"]:
        return jr.poisson(key, expected_electron_events).astype(float)



from ._mrc import (
    read_array_from_mrc as read_array_from_mrc,
    write_image_stack_to_mrc as write_image_stack_to_mrc,
    write_image_to_mrc as write_image_to_mrc,
    write_volume_to_mrc as write_volume_to_mrc,
)
from ._pdb import (
    mmdf_to_atoms as mmdf_to_atoms,
    mmdf_to_topology as mmdf_to_topology,
    read_atoms_from_pdb as read_atoms_from_pdb,
    read_topology_from_pdb as read_topology_from_pdb,
)
from ._starfile import (
    read_starfile as read_starfile,
    write_starfile as write_starfile,
)



import warnings

import cryojax.simulator as cxs
import jax.numpy as jnp
import numpy as np
import pytest
from cryojax.io import read_array_from_mrc
from cryojax.ndimage import (
    cartesian_to_polar,
    compute_binned_powerspectrum,
    irfftn,
    make_frequency_grid,
)
from cryojax.simulator import AstigmaticCTF, EulerAnglePose


try:
    from pycistem.core import CTF as cistemCTF, AnglesAndShifts, Image  # pyright: ignore
except ModuleNotFoundError:
    cistemCTF, AnglesAndShifts, Image = None, None, None

PYCISTEM_WARNING_MESAGE = (
    "Testing against cisTEM is not running because `pycistem` was not "
    "found. Note that `pycistem` cannot be installed on non-linux OS."
)


@pytest.mark.parametrize(
    "defocus1,defocus2,asti_angle,kV,cs,ac,pixel_size",
    [
        (12000.0, 12000.0, 0.0, 300.0, 2.7, 0.07, 1.0),
        (12000.0, 12000.0, 0.0, 200.0, 0.01, 0.12, 1.3),
        (1200.0, 1200.0, 0.0, 300.0, 2.7, 0.07, 1.5),
        (24000.0, 12000.0, 30.0, 300.0, 2.7, 0.07, 0.9),
        (24000.0, 24000.0, 0.0, 300.0, 2.7, 0.07, 2.0),
        (9000.0, 7000.0, 180.0, 300.0, 2.7, 0.07, 1.0),
        (12000.0, 9000.0, 0.0, 200.0, 2.7, 0.07, 0.9),
        (12000.0, 12000.0, 60.0, 200.0, 2.7, 0.02, 0.75),
        (12000.0, 3895.0, 45.0, 200.0, 2.7, 0.07, 2.2),
    ],
)
def test_ctf_with_cistem(defocus1, defocus2, asti_angle, kV, cs, ac, pixel_size):
    """Test CTF model against cisTEM.

    Modified from https://github.com/jojoelfe/contrasttransferfunction"""
    if cistemCTF is not None:
        shape = (512, 512)
        freqs = make_frequency_grid(shape, pixel_size)
        k_sqr, theta = cartesian_to_polar(freqs, square=True)
        # Compute cryojax CTF
        optics = AstigmaticCTF(
            defocus_in_angstroms=(defocus1 + defocus2) / 2,
            astigmatism_in_angstroms=defocus1 - defocus2,
            astigmatism_angle=asti_angle,
            spherical_aberration_in_mm=cs,
        )
        ctf = jnp.array(
            optics(freqs, voltage_in_kilovolts=kV, amplitude_contrast_ratio=ac)
        )
        # Compute cisTEM CTF
        cisTEM_optics = cistemCTF(
            kV=kV,
            cs=cs,
            ac=ac,
            defocus1=defocus1,
            defocus2=defocus2,
            astig_angle=asti_angle,
            pixel_size=pixel_size,
        )
        cisTEM_ctf = np.vectorize(
            lambda k_sqr, theta: cisTEM_optics.Evaluate(k_sqr, theta)
        )(k_sqr.ravel() * pixel_size**2, theta.ravel()).reshape(freqs.shape[0:2])
        # cisTEM_ctf[0, 0] = 0.0

        # Compute cryojax and cisTEM power spectrum
        radial_freqs = jnp.linalg.norm(freqs, axis=-1)
        spectrum1D, _ = compute_binned_powerspectrum(
            ctf, radial_freqs, pixel_size, maximum_frequency=1 / (2 * pixel_size)
        )
        cisTEM_spectrum1D, _ = compute_binned_powerspectrum(
            cisTEM_ctf, radial_freqs, pixel_size, maximum_frequency=1 / (2 * pixel_size)
        )

        np.testing.assert_allclose(ctf, cisTEM_ctf, atol=5e-2)
        np.testing.assert_allclose(spectrum1D, cisTEM_spectrum1D, atol=5e-3)


@pytest.mark.parametrize(
    "phi, theta, psi",
    [(10.0, 90.0, 170.0), (10.0, 80.0, -20.0), (-1.2, 90.5, 67.0), (-50.0, 62.0, -21.0)],
)
def test_euler_matrix_with_cistem(phi, theta, psi):
    """Test zyz rotation matrix"""
    # Hard code zyz rotation matrix from cisTEM convention
    phi_in_rad, theta_in_rad, psi_in_rad = [
        np.deg2rad(angle) for angle in [phi, theta, psi]
    ]
    matrix = np.zeros((3, 3))
    cos_phi = np.cos(phi_in_rad)
    sin_phi = np.sin(phi_in_rad)
    cos_theta = np.cos(theta_in_rad)
    sin_theta = np.sin(theta_in_rad)
    cos_psi = np.cos(psi_in_rad)
    sin_psi = np.sin(psi_in_rad)
    matrix[0, 0] = cos_phi * cos_theta * cos_psi - sin_phi * sin_psi
    matrix[0, 1] = sin_phi * cos_theta * cos_psi + cos_phi * sin_psi
    matrix[0, 2] = -sin_theta * cos_psi
    matrix[1, 0] = -cos_phi * cos_theta * sin_psi - sin_phi * cos_psi
    matrix[1, 1] = -sin_phi * cos_theta * sin_psi + cos_phi * cos_psi
    matrix[1, 2] = sin_theta * sin_psi
    matrix[2, 0] = sin_theta * cos_phi
    matrix[2, 1] = sin_theta * sin_phi
    matrix[2, 2] = cos_theta
    # Generate rotation that matches this rotation matrix
    pose = EulerAnglePose(phi_angle=-phi, theta_angle=-theta, psi_angle=-psi)
    np.testing.assert_allclose(pose.rotation.as_matrix(), matrix, atol=1e-12)


@pytest.mark.parametrize(
    "phi, theta, psi",
    [(10.0, 90.0, 170.0)],
)
def test_compute_projection_with_cistem(
    phi,
    theta,
    psi,
    sample_mrc_path,
):
    if AnglesAndShifts is not None:
        # cryojax
        real_voxel_grid, voxel_size = read_array_from_mrc(
            sample_mrc_path, loads_grid_spacing=True
        )
        volume = cxs.FourierVoxelGridVolume.from_real_voxel_grid(real_voxel_grid)
        pose = cxs.EulerAnglePose(phi_angle=-phi, theta_angle=-theta, psi_angle=-psi)
        projection_method = cxs.FourierSliceExtraction()
        box_size = volume.shape[0]
        image_config = cxs.BasicImageConfig((box_size, box_size), voxel_size, 300.0)
        cryojax_projection = irfftn(
            (
                projection_method.integrate(
                    volume.rotate_to_pose(pose), image_config, outputs_real_space=False
                )
                / voxel_size
            )
            .at[0, 0]
            .set(0.0 + 0.0j)
            / np.sqrt(np.prod(image_config.shape)),
            s=image_config.padded_shape,
        )
        # pycistem
        pycistem_volume = _load_pycistem_template(sample_mrc_path, box_size)
        pycistem_angles = AnglesAndShifts()
        pycistem_angles.Init(phi, theta, psi, 0.0, 0.0)
        pycistem_model = _compute_projection(pycistem_volume, pycistem_angles, box_size)
        pycistem_projection = np.asarray(pycistem_model.real_values)

        np.testing.assert_allclose(cryojax_projection, pycistem_projection, atol=1e-5)
    else:
        warnings.warn(PYCISTEM_WARNING_MESAGE)


def _load_pycistem_template(filename, box_size):
    """Load pycistem template in fourier space."""
    volume = Image()  # type: ignore
    volume.QuickAndDirtyReadSlices(filename, 1, box_size)
    volume.ForwardFFT(True)
    volume.ZeroCentralPixel()
    volume.SwapRealSpaceQuadrants()

    return volume


def _compute_projection(volume, angles, box_size):
    """Compute scattering projection of pycistem volume."""
    projection = Image()  # type: ignore
    projection.Allocate(box_size, box_size, False)

    volume.ExtractSlice(projection, angles, 1.0, False)
    projection.PhaseShift(angles.ReturnShiftX(), angles.ReturnShiftY(), 0.0)

    projection.SwapRealSpaceQuadrants()
    projection.BackwardFFT()

    return projection



from collections import namedtuple


Parkhurst2024PowerConstants = namedtuple(
    "Parkhurst2024PowerConstants", ["a_1", "a_2", "s_1", "s_2", "m"]
)

PARKHURST2024_POWER_CONSTANTS = Parkhurst2024PowerConstants(
    a_1=0.199, a_2=0.801, s_1=0.731, s_2=0.081, m=1 / 2.88
)
"""Constants for used to define the power envelope from Parkhurst et al. (2024)

**References:**

- Parkhurst, James M., et al. "Computational models of amorphous ice for accurate
simulation of cryo-EM images of biological samples." Ultramicroscopy 256 (2024): 113882.
"""



"""The main search loop for the grid search."""

import math
import warnings
from collections.abc import Callable
from typing import Any

import equinox as eqx
import equinox.internal as eqxi
import jax
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import Array, PyTree

from ..._internal import fori_loop_tqdm_decorator
from .custom_types import PyTreeGrid, PyTreeGridPoint
from .pytree_manipulation import (
    tree_grid_shape,
    tree_grid_take,
    tree_grid_unravel_index,
)
from .search_method import AbstractGridSearchMethod


@eqx.filter_jit
def run_grid_search(
    fn: Callable[[PyTreeGridPoint, Any], Array],
    method: AbstractGridSearchMethod,
    tree_grid: PyTreeGrid,
    args: Any,
    *,
    is_leaf: Callable[[Any], bool] | None = None,
    progress_bar: bool = False,
    total_updates: int = 20,
) -> PyTree[Any]:
    """Run a grid search to minimize the function `fn`.

    !!! question "What is a `tree_grid`?"

        For the grid search, we represent the grid as an arbitrary
        pytree whose leaves are JAX arrays with a leading dimension.
        For a particular leaf, its leading dimension indexes a set
        grid points. The entire grid is then the cartesian product
        of the grid points of all of its leaves.

    !!! warning

        A `tree_grid` can only have leaves that are JAX arrays of
        grid points and `None`. It is difficult to precisely check this
        condition even with a run-time type checker, so breaking it may
        result in unhelpful errors.

    To learn more, see the `tree_grid` manipulation routines [`tree_grid_shape`][] and
    [`tree_grid_take`][].

    **Arguments:**

    - `fn`: The function we would like to minimize with grid search. This
            should be evaluated at arguments `fn(y, args)`, where `y` is a
            particular grid point of `tree_grid`. The value returned by `fn`
            must be compatible with the respective `method`.
    - `method`: An interface that specifies what we would like to do with
                each evaluation of `fn`.
    - `tree_grid`: The grid as a pytree. Importantly, its leaves can only be JAX
                   arrays with leading dimensions and `None`.
    - `args`: Arguments passed to `fn`, as `fn(y, args)`.
    - `is_leaf`: As [`jax.tree_util.tree_flatten`](https://jax.readthedocs.io/en/latest/_autosummary/jax.tree_util.tree_flatten.html).
                 This specifies what is to be treated as a leaf in `tree_grid`.
    - `progress_bar`: Add a [`tqdm`](https://github.com/tqdm/tqdm) progress bar to the
                      search loop.
    - `total_updates`:
        An interval for the number of iterations at which toupdate the
        tqdm progress bar. By default, this is `20`. Ignored if `progress_bar = False`.

    **Returns:**

    Any pytree, as specified by the method `AbstractGridSearchMethod.postprocess`.
    """
    warnings.warn(
        "The `cryojax.jax_util` grid search API is deprecated "
        "and will be removed in cryoJAX 0.6.0. Instead, use the "
        "library `brutax` (https://github.com/michael-0brien/brutax).",
        category=FutureWarning,
        stacklevel=2,
    )
    # Evaluate the shape and dtype of the output of `fn` using
    # eqx.filter_closure_convert.
    test_tree_grid_point = tree_grid_take(
        tree_grid,
        tree_grid_unravel_index(0, tree_grid, is_leaf=is_leaf),
    )
    fn = eqx.filter_closure_convert(fn, test_tree_grid_point, args)
    f_struct = jtu.tree_map(
        lambda x: x.value,
        jtu.tree_map(eqxi.Static, fn.out_struct),  # type: ignore
        is_leaf=lambda x: isinstance(x, eqxi.Static),
    )
    # Get the initial state of the search method
    init_state = method.init(tree_grid, f_struct, is_leaf=is_leaf)
    dynamic_init_state, static_state = eqx.partition(init_state, eqx.is_array)
    # Finally, build the loop
    init_carry = (dynamic_init_state, tree_grid)

    def brute_force_body_fun(iteration_index, carry):
        dynamic_state, tree_grid = carry
        state = eqx.combine(static_state, dynamic_state)
        tree_grid_point = tree_grid_take(
            tree_grid,
            tree_grid_unravel_index(iteration_index, tree_grid, is_leaf=is_leaf),
        )
        new_state = method.update(fn, tree_grid_point, args, state, iteration_index)
        new_dynamic_state, new_static_state = eqx.partition(new_state, eqx.is_array)
        assert eqx.tree_equal(static_state, new_static_state) is True
        return new_dynamic_state, tree_grid

    def batched_body_fun(iteration_index, carry):
        dynamic_state, tree_grid = carry
        state = eqx.combine(static_state, dynamic_state)
        raveled_grid_index_batch = jnp.linspace(
            iteration_index * method.batch_size,
            (iteration_index + 1) * method.batch_size - 1,
            method.batch_size,  # type: ignore
            dtype=jnp.int32,
        )
        tree_grid_points = tree_grid_take(
            tree_grid,
            tree_grid_unravel_index(raveled_grid_index_batch, tree_grid, is_leaf=is_leaf),
        )
        new_state = method.batch_update(
            fn, tree_grid_points, args, state, raveled_grid_index_batch
        )
        new_dynamic_state, new_static_state = eqx.partition(new_state, eqx.is_array)
        assert eqx.tree_equal(static_state, new_static_state) is True
        return new_dynamic_state, tree_grid

    # Get the number of iterations of the loop (the size of the grid)
    grid_size = math.prod(tree_grid_shape(tree_grid, is_leaf=is_leaf))
    if method.batch_size is None:
        n_iterations = grid_size
        body_fun = brute_force_body_fun
    else:
        n_iterations = grid_size // method.batch_size
        body_fun = batched_body_fun
    # Run and unpack results
    if progress_bar:
        print_every = n_iterations // total_updates
        if print_every == 0:
            raise ValueError(
                "The number of progress bar updates is greater than the "
                "number of iterations. Try decreasing `total_updates` "
                "or setting it to `None`."
            )
        try:
            body_fun = fori_loop_tqdm_decorator(n_iterations, print_every)(body_fun)
        except ModuleNotFoundError as err:
            raise ModuleNotFoundError(
                "Could not use `progress_bar = True` because `tqdm` was not installed."
            ) from err
    final_carry = jax.lax.fori_loop(0, n_iterations, body_fun, init_carry)
    if method.batch_size is not None and grid_size % method.batch_size != 0:
        final_carry = jax.lax.fori_loop(
            n_iterations,
            n_iterations + (grid_size % method.batch_size),
            batched_body_fun,
            final_carry,
        )
    dynamic_final_state, _ = final_carry
    final_state = eqx.combine(static_state, dynamic_final_state)
    # Return the solution
    solution = method.postprocess(tree_grid, final_state, f_struct, is_leaf=is_leaf)
    return solution



import cryojax.ndimage as im
import cryojax.simulator as cxs
import jax
import jax.numpy as jnp
import numpy as np
import pytest
from cryojax.io import read_array_from_mrc
from cryojax.ndimage import make_coordinate_grid, make_frequency_grid


@pytest.fixture
def voxel_info(sample_mrc_path):
    return read_array_from_mrc(sample_mrc_path, loads_grid_spacing=True)


@pytest.fixture
def voxel_volume(voxel_info):
    return cxs.FourierVoxelGridVolume.from_real_voxel_grid(voxel_info[0], pad_scale=1.3)


@pytest.fixture
def voxel_size(voxel_info):
    return voxel_info[1]


@pytest.fixture
def basic_config(voxel_volume, voxel_size):
    shape = voxel_volume.shape[0:2]
    return cxs.BasicImageConfig(
        shape=shape,
        pixel_size=voxel_size,
        voltage_in_kilovolts=300.0,
    )


def test_mask_2d_running():
    classes = [
        im.SincCorrectionMask,
        im.SquareCosineMask,
        im.CircularCosineMask,
        im.Cylindrical2DCosineMask,
        im.Rectangular2DCosineMask,
    ]
    kwargs = [
        dict(),
        dict(side_length=5, rolloff_width=2),
        dict(radius=5, rolloff_width=2),
        dict(radius=5, rolloff_width=2, length=5, rotation_angle=2.0),
        dict(x_width=5, y_width=5, rolloff_width=2, rotation_angle=2.0),
    ]
    coordinate_grid = make_coordinate_grid((10, 10))
    image = jnp.zeros((10, 10))
    for i, cls in enumerate(classes):
        mask = cls(coordinate_grid, **kwargs[i])
        _ = mask.get()
        _ = mask(image)


def test_mask_3d_running():
    classes = [im.SincCorrectionMask, im.SphericalCosineMask, im.Rectangular3DCosineMask]
    kwargs = [
        dict(),
        dict(radius=5, rolloff_width=2),
        dict(x_width=5, y_width=5, z_width=5, rolloff_width=2),
    ]
    coordinate_grid = make_coordinate_grid((10, 10, 10))
    image = jnp.zeros((10, 10, 10))
    for i, cls in enumerate(classes):
        mask = cls(coordinate_grid, **kwargs[i])
        _ = mask.get()
        _ = mask(image)


def test_filter_running():
    classes = [im.LowpassFilter, im.HighpassFilter]
    kwargs = [dict(), dict()]
    frequency_grid_2d, fourier_image_2d = (
        make_frequency_grid((10, 10)),
        jnp.zeros((10, 10 // 2 + 1)),
    )
    frequency_grid_3d, fourier_image_3d = (
        make_frequency_grid((10, 10, 10)),
        jnp.zeros((10, 10, 10 // 2 + 1)),
    )
    for i, cls in enumerate(classes):
        f_2d = cls(frequency_grid_2d, **kwargs[i])
        _ = f_2d.get()
        _ = f_2d(fourier_image_2d)
        f_3d = cls(frequency_grid_3d, **kwargs[i])
        _ = f_3d.get()
        _ = f_3d(fourier_image_3d)


def test_custom_filter_and_mask_initialization():
    classes = [im.CustomFilter, im.CustomMask]
    array = jnp.zeros((10, 10))
    for cls in classes:
        _ = cls(array)


@pytest.mark.parametrize(
    "image_shape, filter_shape, mode, square",
    (
        ((10, 10), None, "linear", False),
        ((2, 10, 10), None, "linear", False),
        ((2, 10, 10), (9, 9), "linear", False),
        ((2, 10, 10), (11, 11), "linear", False),
        ((2, 10, 10), None, "nearest", False),
        ((2, 10, 10), None, "linear", True),
    ),
)
def test_whitening_filter(image_shape, filter_shape, mode, square):
    rng_key = jax.random.key(1234)
    image = jax.random.normal(rng_key, image_shape)
    f = im.WhiteningFilter(
        image, shape=filter_shape, interpolation_mode=mode, outputs_squared=square
    )
    _ = f.get()


@pytest.mark.parametrize("use_rfft", [False])
def test_rotation_fn(basic_config, voxel_volume, use_rfft):
    rotation_angle = 35.0
    pose_norot = cxs.EulerAnglePose(theta_angle=90.0, psi_angle=0.0)
    pose_ref = cxs.EulerAnglePose(theta_angle=90.0, psi_angle=rotation_angle)
    image_model_norot = cxs.make_image_model(voxel_volume, basic_config, pose=pose_norot)
    image_model_ref = cxs.make_image_model(voxel_volume, basic_config, pose=pose_ref)

    if use_rfft:
        grid = basic_config.get_frequency_grid(physical=False, padding=True)
    else:
        grid = basic_config.get_frequency_grid(physical=False, full=True, padding=True)
    rotation_fn = im.RotateFFT(rotation_angle, grid)

    image_norot = image_model_norot.raw_simulate()
    image_ref = image_model_ref.raw_simulate()
    if use_rfft:
        shape = basic_config.padded_shape
        image_rot = im.irfftn(rotation_fn(im.rfftn(image_norot)), s=shape)
    else:
        image_rot = im.ifftn(rotation_fn(im.fftn(image_norot))).real

    corr = _get_correlation(image_ref, image_rot)
    np.testing.assert_allclose(corr.item(), 1.0, atol=1e-1)


@pytest.mark.parametrize("use_rfft", [True, False])
def test_translation_fn(basic_config, voxel_volume, use_rfft):
    pose_notranslate = cxs.EulerAnglePose()
    pose_translate = cxs.EulerAnglePose(
        offset_x_in_angstroms=50.0, offset_y_in_angstroms=-30.0
    )

    image_model_notrans = cxs.make_image_model(
        voxel_volume,
        basic_config,
        pose=pose_notranslate,
        translate_mode="none",
    )

    image_model_ref = cxs.make_image_model(
        voxel_volume,
        basic_config,
        pose=pose_translate,
    )
    if use_rfft:
        grid = basic_config.get_frequency_grid(physical=True)
    else:
        grid = basic_config.get_frequency_grid(physical=True, full=True)

    shift_fn = im.PhaseShiftFFT(
        offset=jnp.array([50.0, -30.0]),
        frequency_grid=grid,
    )

    image_notrans = image_model_notrans.simulate()
    image_ref = image_model_ref.simulate()
    if use_rfft:
        image_trans = im.irfftn(
            shift_fn(im.rfftn(image_notrans)), s=basic_config.padded_shape
        )
    else:
        image_trans = im.ifftn(shift_fn(im.fftn(image_notrans))).real

    np.testing.assert_allclose(image_ref, image_trans)


def _get_correlation(im1, im2):
    return jnp.abs(jnp.sum(im1 * im2)) / (jnp.linalg.norm(im1) * jnp.linalg.norm(im2))



from .base_particle_dataset import (
    AbstractParticleDataset as AbstractParticleDataset,
    AbstractParticleParameterFile as AbstractParticleParameterFile,
)
from .particle_simulation import simulate_particle_stack as simulate_particle_stack
from .relion import (
    AbstractParticleStarFile as AbstractParticleStarFile,
    RelionParticleDataset as RelionParticleDataset,
    RelionParticleParameterFile as RelionParticleParameterFile,
)



from typing import Any, ClassVar, Literal
from typing_extensions import override

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Complex, Float

from ...jax_util import FloatLike, NDArrayLike
from ...rotations import SO2
from .._fourier_utils import enforce_rfftn_self_conjugates
from .._map_coordinates import map_coordinates
from .._operators import FourierPhaseShifts
from ._base_transform import AbstractImageTransform


class PhaseShiftFFT(AbstractImageTransform, strict=True):
    """Apply a phase shift to an image in Fourier space, effectively
    applying an in-plane shift to the image in real space. Only square
    images are supported.

    !!! example "Apply a translation in real-space"

        ```python

        from cryojax.ndimage import PhaseShiftFFT, rfftn, irfftn

        offset_in_angstroms = jnp.array([50.0, -30.0])
        frequency_grid = ... # in angstroms
        fft = rfftn(...) # e.g., fft of a real 2D image

        shift_fn = PhaseShiftFFT(
            offset=offset_in_angstroms, frequency_grid=frequency_grid
        )

        shifted_fft = shift_fn(fft)
        shifted_image = irfftn(shifted_image_fft)
        ```
    """

    translation_operator: Complex[Array, "_ _ _"] | Complex[Array, "_ _"]
    is_rfft: bool = eqx.field(static=True)

    is_real_space: ClassVar[bool] = False

    def __init__(
        self,
        offset: Float[NDArrayLike, "2"] | Float[NDArrayLike, "3"],
        frequency_grid: Float[NDArrayLike, "_ _ 2"] | Float[NDArrayLike, "_ _ _ 3"],
    ):
        """**Arguments:**

        - `offset`: The offset by which to shift the image, in pixels or angstroms.
        - `frequency_grid`: The frequency grid in pixels or angstroms.
        """
        if _is_square_rfft_grid(frequency_grid):
            # It's worth noting that this condition is breakable and rfftn/fftn
            # inference can fail! One can pass a grid meant for use with fftn
            # with these exact shapes.
            self.is_rfft = True
        elif _is_square_fft_grid(frequency_grid):
            self.is_rfft = False
        else:
            raise ValueError(
                "The `frequency_grid` argument to `PhaseShiftFFT` did not have a valid "
                f"shape {frequency_grid.shape}. `PhaseShiftFFT` only supports square "
                "images as input; you may have passed a grid that does not correspond "
                "to a square image."
            )
        compute_operator = FourierPhaseShifts(offset)
        self.translation_operator = compute_operator(
            jnp.asarray(frequency_grid, dtype=float)
        )

    @override
    def __call__(
        self, image: Complex[Array, "y_dim x_dim"] | Complex[Array, "z_dim y_dim x_dim"]
    ) -> Complex[Array, "y_dim x_dim"] | Complex[Array, "z_dim y_dim x_dim"]:
        """Apply the phase shift to the input image in Fourier space.

        **Arguments:**

        - `image`:
            The input image in Fourier space.

        **Returns:**

        The phase shifted image in Fourier space.
        """
        if image.shape != self.translation_operator.shape:
            raise ValueError(
                "The image passed to `PhaseShiftFFT` did not have a valid "
                f"shape. The shape of the image was {image.shape}, "
                "but that of the translation operator was "
                f"{self.translation_operator.shape}."
            )
        if self.is_rfft:
            ndim, dim = self.translation_operator.ndim, self.translation_operator.shape[0]
            shape = tuple(ndim * [dim])
            image = enforce_rfftn_self_conjugates(
                image,
                shape,  # pyright: ignore[reportArgumentType]
                includes_dc=False,
                mode="zero",
            )

        return image * self.translation_operator


class RotateFFT(AbstractImageTransform, strict=True):
    """Rotate an image in Fourier space using interpolation.
    Only square images are supported.

    !!! example

        ```python

        from cryojax.ndimage import RotateFFT, fftn, ifftn

        frequency_grid = ... # in pixels
        image = ... # e.g., a real 2D image

        rotation_fn = RotateFFT(
            rotation_angle=45.0, frequency_grid=frequency_grid
        )

        rotated_image_fft = rotation_fn(fftn(image))
        rotated_image = ifftn(rotated_image_fft).real
        ```
    """

    rotation_angle: Float[Array, ""]
    frequency_grid: Float[Array, "_ _ 2"]
    rotation_convention: Literal["frame", "object"]
    map_coordinates_options: dict[str, Any]

    is_rfft: bool = eqx.field(static=True)

    is_real_space: ClassVar[bool] = False

    def __init__(
        self,
        rotation_angle: FloatLike,
        frequency_grid: Float[NDArrayLike, "y_dim x_dim 2"],
        *,
        pixel_size: FloatLike = 1.0,
        rotation_convention: Literal["frame", "object"] = "object",
        map_coordinates_options: dict[str, Any] = {},
    ):
        """
        **Arguments:**

        - `rotation_angle`: The angle by which to rotate the image, in degrees.
        - `frequency_grid`: The frequency grid.
        - `pixel_size`: The pixel size of the `frequency_grid`.
        - `rotation_convention`:
            If `'object'`, the rotation is with respect to the object in the image.
            If `'frame'`, it is with respect to the frame. These are related by transpose.
        - `map_coordinates_options`:
            A dictionary of options passed to [`cryojax.ndimage.map_coordinates`][].
        """
        if rotation_convention not in ["object", "frame"]:
            raise ValueError(
                "Invalid value for the `rotation_convention` argument to `RotateFFT`. "
                f"Found {rotation_convention}, but options are 'object' or 'frame'."
            )
        if _is_square_rfft_grid(frequency_grid, only_2d=True):
            # It's worth noting that this condition is breakable and rfftn/fftn
            # inference can fail! One can pass a grid meant for use with fftn
            # with these exact shapes.
            self.is_rfft = True
        elif _is_square_fft_grid(frequency_grid, only_2d=True):
            self.is_rfft = False
        else:
            raise ValueError(
                "The `frequency_grid` argument to `RotateFFT` did not have a valid "
                f"shape {frequency_grid.shape}. `RotateFFT` only supports square "
                "2D images as input; you may have passed a grid that does not "
                "correspond to a square image, or you may have tried to use `RotateFFT` "
                "with a volume."
            )
        self.rotation_angle = jnp.asarray(rotation_angle, dtype=float)
        self.rotation_convention = rotation_convention
        self.frequency_grid = jnp.asarray(frequency_grid * pixel_size, dtype=float)
        self.map_coordinates_options = map_coordinates_options

    @override
    def __call__(
        self, image: Complex[Array, "y_dim x_dim"]
    ) -> Complex[Array, "y_dim x_dim"]:
        """Rotate the input image in Fourier space.

        **Arguments:**

        `image`:
            The image in Fourier space.

        **Returns:**

        The rotated image in Fourier space.
        """
        if image.shape != self.frequency_grid.shape[0:-1]:
            raise ValueError(
                "The image passed to `RotateFFT` did not have a valid "
                f"shape. The shape of the image was {image.shape}, "
                "but that of the `frequency_grid` was "
                f"{self.frequency_grid.shape}."
            )
        shift_axes = (0,) if self.is_rfft else (0, 1)
        # Shift images and grid so that zero is in center
        fourier_image = image
        fourier_image_c = jnp.fft.fftshift(fourier_image, axes=shift_axes)
        frequency_grid_c = jnp.fft.fftshift(self.frequency_grid, axes=shift_axes)
        # Rotate the grid
        rotation_matrix = _get_rotation_matrix(
            self.rotation_angle, self.rotation_convention
        )
        rotated_grid = frequency_grid_c @ rotation_matrix
        # Interpolate at new coordinates
        logical_grid = _frequencies_to_indices(rotated_grid, self.is_rfft)
        k_x, k_y = jnp.transpose(logical_grid, axes=[2, 0, 1])
        rotated_image_c = map_coordinates(
            fourier_image_c, (k_y, k_x), **self.map_coordinates_options
        )
        # Shift back, ensure that rfft components are real-valued where they
        # should be, and return
        rotated_image = jnp.fft.ifftshift(rotated_image_c, axes=shift_axes)
        if self.is_rfft:
            dim = rotated_image.shape[0]
            rotated_image = enforce_rfftn_self_conjugates(
                rotated_image, (dim, dim), includes_dc=True, mode="real"
            )
        return rotated_image


def _get_rotation_matrix(angle, convention):
    if convention == "object":
        angle *= -1.0
    angle = jnp.deg2rad(angle)
    c, s = jnp.cos(angle), jnp.sin(angle)
    rotation = SO2([c, s])
    return rotation.as_matrix()


def _frequencies_to_indices(freqs, is_rfft):
    dim = freqs.shape[0]
    if is_rfft:
        freqsx, freqsy = freqs[..., 0], freqs[..., 1]
        return jnp.stack([freqsx * dim, freqsy * dim + dim // 2], axis=-1)
    else:
        return dim * freqs + dim // 2


def _is_square_rfft_grid(grid, only_2d: bool = False):
    shape, dim = grid.shape, grid.shape[0]
    shapes_2d = [(dim, dim // 2 + 1, 2), (dim, dim // 2, 2)]
    if only_2d:
        return shape in shapes_2d
    else:
        return shape in [*shapes_2d, (dim, dim, dim // 2, 3), (dim, dim, dim // 2 + 1, 3)]


def _is_square_fft_grid(grid, only_2d: bool = False):
    shape, dim = grid.shape, grid.shape[0]
    if only_2d:
        return shape == (dim, dim, 2)
    else:
        return shape in [(dim, dim, 2), (dim, dim, dim, 3)]



"""
Masks to apply to images in real space.
"""

import abc
import functools
import operator
from typing import ClassVar
from typing_extensions import override

import jax
import jax.numpy as jnp
from jaxtyping import Array, Float

from ...jax_util import FloatLike, NDArrayLike
from ._base_transform import AbstractImageTransform


class AbstractMask(AbstractImageTransform, strict=True):
    """Base class for computing and applying an image mask."""

    is_real_space: ClassVar[bool] = True

    @abc.abstractmethod
    def get(self) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        raise NotImplementedError

    def __call__(
        self, image: Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]
    ) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        return image * self.get()


class CustomMask(AbstractMask, strict=True):
    """Pass a custom mask as an array."""

    array: Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]

    def __init__(
        self, mask_array: Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]
    ):
        self.array = mask_array

    @override
    def get(self) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        return self.array


class CircularCosineMask(AbstractMask, strict=True):
    """Apply a circular mask to an image with a cosine
    soft-edge.
    """

    array: Float[Array, "y_dim x_dim"]

    def __init__(
        self,
        coordinate_grid: Float[Array, "y_dim x_dim 2"],
        radius: FloatLike,
        rolloff_width: FloatLike,
        xy_offset: tuple[float, float] | Float[NDArrayLike, " 2"] = (0.0, 0.0),
    ):
        """**Arguments:**

        - `coordinate_grid`:
            The image coordinates.
        - `radius`:
            The radius of the circular mask.
        - `rolloff_width`:
            The rolloff width of the soft edge.
        """
        self.array = _compute_circular_or_spherical_mask(
            coordinate_grid,
            jnp.asarray(radius),
            jnp.asarray(rolloff_width),
            offset=jnp.asarray(xy_offset),
        )

    @override
    def get(self) -> Float[Array, "y_dim x_dim"]:
        return self.array


class SphericalCosineMask(AbstractMask, strict=True):
    """Apply a spherical mask to a volume with a cosine
    soft-edge.
    """

    array: Float[Array, "z_dim y_dim x_dim"]

    def __init__(
        self,
        coordinate_grid: Float[Array, "z_dim y_dim x_dim 3"],
        radius: FloatLike,
        rolloff_width: FloatLike,
    ):
        """**Arguments:**

        - `coordinate_grid`:
            The volume coordinates.
        - `radius`:
            The radius of the spherical mask.
        - `rolloff_width`:
            The rolloff width of the soft edge.
        """
        self.array = _compute_circular_or_spherical_mask(
            coordinate_grid,
            jnp.asarray(radius),
            jnp.asarray(rolloff_width),
            offset=None,
        )

    @override
    def get(self) -> Float[Array, "z_dim y_dim x_dim"]:
        return self.array


class SquareCosineMask(AbstractMask, strict=True):
    """Apply a square mask to an image with a cosine
    soft-edge.
    """

    array: Float[Array, "y_dim x_dim"]

    def __init__(
        self,
        coordinate_grid: Float[Array, "y_dim x_dim 2"],
        side_length: FloatLike,
        rolloff_width: FloatLike,
    ):
        """**Arguments:**

        - `coordinate_grid`:
            The image coordinates.
        - `side_length`:
            The side length of the square.
        - `rolloff_width`:
            The rolloff width of the soft edge.
        """
        self.array = _compute_square_mask(
            coordinate_grid, jnp.asarray(side_length), jnp.asarray(rolloff_width)
        )

    @override
    def get(self) -> Float[Array, "y_dim x_dim"]:
        return self.array


class Cylindrical2DCosineMask(AbstractMask, strict=True):
    """Apply a cylindrical mask to an image with a cosine
    soft-edge. This implements an infinite in-plane cylinder,
    rotated at a given angle.
    """

    array: Float[Array, "y_dim x_dim"]

    def __init__(
        self,
        coordinate_grid: Float[Array, "y_dim x_dim 2"],
        radius: FloatLike,
        rolloff_width: FloatLike,
        rotation_angle: FloatLike = 0.0,
        length: FloatLike | None = None,
    ):
        """**Arguments:**

        - `coordinate_grid`:
            The image coordinates.
        - `radius`:
            The radius of the cylinder.
        - `rolloff_width`:
            The rolloff width of the soft edge.
        - `rotation_angle`:
            The in-plane rotation angle of the cylinder in degrees. By default,
            `0.0`.
        - `length`:
            The length of the cylinder. If `None`, do not mask the cylinder length-wise.
        """
        length = None if length is None else jnp.asarray(length)
        if length is None:
            self.array = _compute_cylindrical_mask_2d_without_length(
                coordinate_grid,
                jnp.asarray(radius),
                jnp.asarray(rotation_angle),
                jnp.asarray(rolloff_width),
            )
        else:
            self.array = _compute_cylindrical_mask_2d_with_length(
                coordinate_grid,
                jnp.asarray(radius),
                length,
                jnp.asarray(rotation_angle),
                jnp.asarray(rolloff_width),
            )

    @override
    def get(self) -> Float[Array, "y_dim x_dim"]:
        return self.array


class Rectangular2DCosineMask(AbstractMask, strict=True):
    """Apply a rectangular mask in 2D to an image with a cosine
    soft-edge. Optionally, rotate the rectangle by an angle.
    """

    array: Float[Array, "y_dim x_dim"]

    def __init__(
        self,
        coordinate_grid: Float[Array, "y_dim x_dim 2"],
        x_width: FloatLike,
        y_width: FloatLike,
        rolloff_width: FloatLike,
        rotation_angle: FloatLike = 0.0,
    ):
        """**Arguments:**

        - `coordinate_grid`:
            The image coordinates.
        - `x_width`:
            The width of the rectangle along the x-axis.
        - `y_width`:
            The width of the rectangle along the y-axis.
        - `rolloff_width`:
            The rolloff width of the soft edge.
        - `rotation_angle`:
            The in-plane rotation angle of the rectangle in degrees. By default,
            `0.0`.
        """
        self.array = _compute_cylindrical_mask_2d_with_length(
            coordinate_grid,
            jnp.asarray(y_width / 2),
            jnp.asarray(x_width),
            jnp.asarray(rotation_angle),
            jnp.asarray(rolloff_width),
        )

    @override
    def get(self) -> Float[Array, "y_dim x_dim"]:
        return self.array


class Rectangular3DCosineMask(AbstractMask, strict=True):
    """Apply a rectangular mask to a volume with a cosine
    soft-edge.
    """

    array: Float[Array, "z_dim y_dim x_dim"]

    def __init__(
        self,
        coordinate_grid: Float[Array, "z_dim y_dim x_dim 3"],
        x_width: FloatLike,
        y_width: FloatLike,
        z_width: FloatLike,
        rolloff_width: FloatLike,
    ):
        """**Arguments:**

        - `coordinate_grid`:
            The volume coordinates.
        - `x_width`:
            The width of the rectangle along the x-axis.
        - `y_width`:
            The width of the rectangle along the y-axis.
        - `z_width`:
            The width of the rectangle along the z-axis.
        - `rolloff_width`:
            The rolloff width of the soft edge.
        """
        self.array = _compute_rectangular_mask_3d(
            coordinate_grid,
            jnp.asarray(x_width),
            jnp.asarray(y_width),
            jnp.asarray(z_width),
            jnp.asarray(rolloff_width),
        )

    @override
    def get(self) -> Float[Array, "z_dim y_dim x_dim"]:
        return self.array


class SincCorrectionMask(AbstractMask, strict=True):
    """Divide an image or volume by a 2D or 3D rectangular
    squared sinc function computed on the unit box. This is used
    for correcting scaling in [`cryojax.simulator.FourierSliceExtraction`][].

    Linear interpolation in the Fourier domain can be thought of as
    a convolution with a triangular kernel, whose Fourier
    transform pair is the squared sinc. This mask
    acts to deconvolve this function with a division in real-space.
    """

    array: Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]

    def __init__(
        self,
        coordinate_grid_in_pixels: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ):
        """**Arguments:**

        - `coordinate_grid_in_pixels`:
            The image or volume coordinates. This should be
            generated with `make_coordinate_grid(shape)`, *not*
            `make_coordinate_grid(shape, grid_spacing)`.
        """
        ndim = coordinate_grid_in_pixels.ndim - 1
        box_dims = coordinate_grid_in_pixels.shape[0:ndim][::-1]
        self.array = (
            jax.lax.reciprocal(
                functools.reduce(
                    operator.mul,
                    [
                        jnp.sinc(coordinate_grid_in_pixels[..., i] / box_dims[i])
                        for i in range(ndim)
                    ],
                )
            )
            ** 2
        )

    @override
    def get(self) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        return self.array


def _compute_circular_or_spherical_mask(
    coordinate_grid: Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"],
    radius: Float[Array, ""],
    rolloff_width: Float[Array, ""],
    offset: Float[Array, "2"] | Float[Array, "3"] | None = None,
) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
    if offset is not None:
        coordinate_grid -= offset[None, None, :]
    radial_coordinate_grid = jnp.linalg.norm(coordinate_grid, axis=-1)

    def compute_mask_at_coordinate(radial_coordinate):
        return jnp.where(
            radial_coordinate <= radius,
            1.0,
            jnp.where(
                radial_coordinate > radius + rolloff_width,
                0.0,
                0.5
                * (1 + jnp.cos(jnp.pi * (radial_coordinate - radius) / rolloff_width)),
            ),
        )

    compute_mask = (
        jax.vmap(jax.vmap(compute_mask_at_coordinate))
        if radial_coordinate_grid.ndim == 2
        else jax.vmap(jax.vmap(jax.vmap(compute_mask_at_coordinate)))
    )

    return compute_mask(radial_coordinate_grid)


def _compute_square_mask(
    coordinate_grid: Float[Array, "y_dim x_dim 2"],
    side_length: Float[Array, ""],
    rolloff_width: Float[Array, ""],
) -> Float[Array, "y_dim x_dim"]:
    is_in_square_fn = lambda abs_x, abs_y, s: jnp.logical_and(
        abs_x <= s / 2, abs_y <= s / 2
    )
    is_in_edge_fn = lambda abs_x_or_y, s, w: jnp.logical_and(
        abs_x_or_y > s / 2, abs_x_or_y < s / 2 + w
    )
    compute_edge_fn = lambda abs_x_or_y, s, w: 0.5 * (
        1 + jnp.cos(jnp.pi * (abs_x_or_y - s / 2) / w)
    )

    def compute_mask_at_coordinate(coordinate):
        x, y = coordinate
        abs_x, abs_y = jnp.abs(x), jnp.abs(y)
        # Check coordinate is in either the square of the unmasked region
        is_in_unmasked_square = is_in_square_fn(abs_x, abs_y, side_length)
        # ... or the square of the unmasked region, plus the rolloff width
        # of the soft edge
        is_in_unmasked_plus_soft_edge_square = is_in_square_fn(
            abs_x, abs_y, side_length + 2 * rolloff_width
        )
        # Next, compute where (if anywhere) the coordinate is in the soft edge
        # region
        is_in_edge_x = is_in_edge_fn(abs_x, side_length, rolloff_width)
        is_in_edge_y = is_in_edge_fn(abs_y, side_length, rolloff_width)
        # Compute the soft edges
        edge_x, edge_y = (
            compute_edge_fn(abs_x, side_length, rolloff_width),
            compute_edge_fn(abs_y, side_length, rolloff_width),
        )

        return jnp.where(
            is_in_unmasked_square,
            1.0,
            jnp.where(
                is_in_unmasked_plus_soft_edge_square,
                jnp.where(
                    jnp.logical_and(is_in_edge_x, is_in_edge_y),
                    edge_x * edge_y,
                    jnp.where(is_in_edge_x, edge_x, edge_y),
                ),
                0.0,
            ),
        )

    compute_mask = jax.vmap(jax.vmap(compute_mask_at_coordinate))

    return compute_mask(coordinate_grid)


def _compute_cylindrical_mask_2d_without_length(
    coordinate_grid: Float[Array, "y_dim x_dim 2"],
    radius: Float[Array, ""],
    angle: Float[Array, ""],
    rolloff_width: Float[Array, ""],
) -> Float[Array, "y_dim x_dim"]:
    # Compute rotated radial coordinate grid
    angle_in_radians = jnp.deg2rad(angle)
    cos_angle = jnp.cos(angle_in_radians)
    sin_angle = jnp.sin(angle_in_radians)
    x, y = coordinate_grid[..., 0], coordinate_grid[..., 1]
    y_rotated = -x * sin_angle + y * cos_angle
    # ... the mask in 2D is just a rectangle; the cylindrical coordinate is
    # the rotated y coordinate
    cylinder_radial_coordinate_grid = jnp.sqrt(y_rotated**2)

    def compute_mask_at_coordinate(cylinder_radial_coordinate):
        r = cylinder_radial_coordinate
        return jnp.where(
            r < radius,
            1.0,
            jnp.where(
                r < radius + rolloff_width,
                0.5 * (1 + jnp.cos(jnp.pi * (r - radius) / rolloff_width)),
                0.0,
            ),
        )

    compute_mask = jax.vmap(jax.vmap(compute_mask_at_coordinate))

    return compute_mask(cylinder_radial_coordinate_grid)


def _compute_cylindrical_mask_2d_with_length(
    coordinate_grid: Float[Array, "y_dim x_dim 2"],
    radius: Float[Array, ""],
    length: Float[Array, ""],
    angle: Float[Array, ""],
    rolloff_width: Float[Array, ""],
) -> Float[Array, "y_dim x_dim"]:
    # Compute rotated radial coordinate grid
    diameter = 2 * radius
    angle_in_radians = jnp.deg2rad(angle)
    cos_angle = jnp.cos(angle_in_radians)
    sin_angle = jnp.sin(angle_in_radians)
    x, y = coordinate_grid[..., 0], coordinate_grid[..., 1]
    x_rotated, y_rotated = x * cos_angle + y * sin_angle, -x * sin_angle + y * cos_angle
    cylinder_r_coordinate_grid = y_rotated
    cylinder_z_coordinate_grid = x_rotated

    is_in_rect_fn = lambda abs_x, abs_y, s_x, s_y: jnp.logical_and(
        abs_x <= s_x / 2, abs_y <= s_y / 2
    )
    is_in_edge_fn = lambda abs_x_or_y, s, w: jnp.logical_and(
        abs_x_or_y > s / 2, abs_x_or_y < s / 2 + w
    )
    compute_edge_fn = lambda abs_x_or_y, s, w: 0.5 * (
        1 + jnp.cos(jnp.pi * (abs_x_or_y - s / 2) / w)
    )

    def compute_mask_at_coordinate(r, z):
        abs_r, abs_z = jnp.abs(r), jnp.abs(z)
        # Check coordinate is in either the rectangle of the unmasked region
        is_in_unmasked_rect = is_in_rect_fn(abs_r, abs_z, diameter, length)
        # ... or the square of the unmasked region, plus the rolloff width
        # of the soft edge
        is_in_unmasked_plus_soft_edge_rect = is_in_rect_fn(
            abs_r, abs_z, diameter + 2 * rolloff_width, length + 2 * rolloff_width
        )
        # Next, compute where (if anywhere) the coordinate is in the soft edge
        # region
        is_in_edge_r = is_in_edge_fn(abs_r, diameter, rolloff_width)
        is_in_edge_z = is_in_edge_fn(abs_z, length, rolloff_width)
        # Compute the soft edges
        edge_r, edge_z = (
            compute_edge_fn(abs_r, diameter, rolloff_width),
            compute_edge_fn(abs_z, length, rolloff_width),
        )

        return jnp.where(
            is_in_unmasked_rect,
            1.0,
            jnp.where(
                is_in_unmasked_plus_soft_edge_rect,
                jnp.where(
                    jnp.logical_and(is_in_edge_r, is_in_edge_z),
                    edge_r * edge_z,
                    jnp.where(is_in_edge_r, edge_r, edge_z),
                ),
                0.0,
            ),
        )

    compute_mask = jax.vmap(jax.vmap(compute_mask_at_coordinate))

    return compute_mask(cylinder_r_coordinate_grid, cylinder_z_coordinate_grid)


def _compute_rectangular_mask_3d(
    coordinate_grid: Float[Array, "z_dim y_dim x_dim 3"],
    x_width: Float[Array, ""],
    y_width: Float[Array, ""],
    z_width: Float[Array, ""],
    rolloff_width: Float[Array, ""],
):
    is_in_rect_fn = lambda abs_x, abs_y, abs_z, s_x, s_y, s_z: jnp.logical_and(
        jnp.logical_and(abs_x <= s_x / 2, abs_y <= s_y / 2), abs_z <= s_z / 2
    )
    is_in_edge_fn = lambda abs_x_or_y_or_z, s, w: jnp.logical_and(
        abs_x_or_y_or_z > s / 2, abs_x_or_y_or_z < s / 2 + w
    )
    compute_edge_fn = lambda abs_x_or_y_or_z, s, w: 0.5 * (
        1 + jnp.cos(jnp.pi * (abs_x_or_y_or_z - s / 2) / w)
    )

    def compute_mask_at_coordinate(coordinate):
        x, y, z = coordinate
        abs_x, abs_y, abs_z = jnp.abs(x), jnp.abs(y), jnp.abs(z)
        # Check coordinate is in either the rectangular of the unmasked region
        is_in_unmasked_rect = is_in_rect_fn(
            abs_x, abs_y, abs_z, x_width, y_width, z_width
        )
        # ... or the rectangle of the unmasked region, plus the rolloff width
        # of the soft edge
        is_in_unmasked_plus_soft_edge_rect = is_in_rect_fn(
            abs_x,
            abs_y,
            abs_z,
            x_width + 2 * rolloff_width,
            y_width + 2 * rolloff_width,
            z_width + 2 * rolloff_width,
        )
        # Next, compute where (if anywhere) the coordinate is in the soft edge
        # region
        is_in_edge_x = is_in_edge_fn(abs_x, x_width, rolloff_width)
        is_in_edge_y = is_in_edge_fn(abs_y, y_width, rolloff_width)
        is_in_edge_z = is_in_edge_fn(abs_z, z_width, rolloff_width)
        # Compute the soft edges
        edge_x, edge_y, edge_z = (
            compute_edge_fn(abs_x, x_width, rolloff_width),
            compute_edge_fn(abs_y, y_width, rolloff_width),
            compute_edge_fn(abs_z, z_width, rolloff_width),
        )

        return jnp.where(
            is_in_unmasked_rect,
            1.0,
            jnp.where(
                is_in_unmasked_plus_soft_edge_rect,
                jnp.where(
                    jnp.logical_and(
                        jnp.logical_and(is_in_edge_x, is_in_edge_y), is_in_edge_z
                    ),
                    edge_x * edge_y * edge_z,
                    jnp.where(
                        jnp.logical_and(is_in_edge_x, is_in_edge_y),
                        edge_x * edge_y,
                        jnp.where(
                            jnp.logical_and(is_in_edge_x, is_in_edge_z),
                            edge_x * edge_z,
                            jnp.where(
                                jnp.logical_and(is_in_edge_y, is_in_edge_z),
                                edge_y * edge_z,
                                jnp.where(
                                    is_in_edge_x,
                                    edge_x,
                                    jnp.where(is_in_edge_y, edge_y, edge_z),
                                ),
                            ),
                        ),
                    ),
                ),
                0.0,
            ),
        )

    compute_mask = jax.vmap(jax.vmap(jax.vmap(compute_mask_at_coordinate)))
    return compute_mask(coordinate_grid)



"""
Transformations used for reparameterizing cryojax models.
"""

from abc import abstractmethod
from collections.abc import Callable
from typing import Any, Generic, TypeVar
from typing_extensions import override

import equinox as eqx
import jax
from equinox import Module, field
from jaxtyping import PyTree


T = TypeVar("T")


def _resolve_transforms(pytree, resolve_self: bool):
    def f(leaf):
        if isinstance(leaf, AbstractPyTreeTransform):
            return _resolve_transforms(leaf, resolve_self=False).value
        else:
            return leaf

    def is_leaf(x):
        is_transform = isinstance(x, AbstractPyTreeTransform)
        return is_transform and (resolve_self or x is not pytree)

    return jax.tree.map(f, pytree, is_leaf=is_leaf)


def resolve_transforms(pytree: T) -> T:
    """Transforms a pytree whose parameters have entries
    that are `AbstractParameterTransform`s back to its
    original parameter space.

    If `AbstractParameterTransform`s are nested, the innermost
    nodes are unwrapped first. This function is based on the implementation
    in the package `paramax`.
    """

    return _resolve_transforms(pytree, resolve_self=True)


class AbstractPyTreeTransform(Module, Generic[T], strict=True):
    """Base class for a pytree transformation.

    This interface can be used to get custom
    per-parameter behavior in pytrees, as described in the [equinox docs](
    https://docs.kidger.site/equinox/tricks/#custom-per-parameter-behaviour).

    For example, this used to do inference in better geometries or to
    enforce parameter constraints, however it can also be used generally for
    assisting with JAX transformations.

    This is a very simple class. When the class is initialized,
    class fields are stored a transformed parameter space. When
    `transform.value` is called, the fields are taken back to
    the original parameter space.
    """

    @property
    @abstractmethod
    def value(self) -> T:
        """Get the value of the transform."""
        raise NotImplementedError


class CustomTransform(AbstractPyTreeTransform[T], strict=True):
    """This class transforms a pytree of arrays using a custom callable."""

    func_dynamic: PyTree
    func_static: PyTree = field(static=True)
    args: tuple[Any, ...]
    kwargs: dict[str, Any]

    def __init__(
        self,
        func: Callable[..., T],
        *args: Any,
        **kwargs: Any,
    ):
        """**Arguments:**

        - `func`:
            A callable of format `func(*args, **kwargs)`. This may be a
            regular function or an pytree with a `__call__` method, such
            as an `equinox.Module`.
        - `args`:
            Arguments to be passed to `fn` at runtime.
        - `kwargs`:
            Keyword arguments to be passed to `fn` at runtime.
        """
        if not isinstance(func, Callable):
            raise ValueError(
                "Argument `func` must be type `Callable`. If `func` "
                "is an `equinox.Module` or another pytree, it must have "
                "a `__call__` method."
            )
        self.func_dynamic, self.func_static = eqx.partition(func, eqx.is_array)
        self.args = tuple(args)
        self.kwargs = kwargs

    @property
    @override
    def value(self) -> T:
        """The pytree transformed with the custom function `func`."""
        func = eqx.combine(self.func_dynamic, self.func_static)
        return func(*self.args, **self.kwargs)


class StopGradientTransform(AbstractPyTreeTransform[T]):
    """Applies stop gradient to all JAX arrays.

    Based on `NonTrainable` from the package `paramax`.
    """

    pytree_differentiable: T
    pytree_static: T = field(static=True)

    def __init__(self, pytree: T):
        self.pytree_differentiable, self.pytree_static = eqx.partition(
            pytree, eqx.is_array
        )

    @property
    @override
    def value(self) -> T:
        return eqx.combine(
            jax.lax.stop_gradient(self.pytree_differentiable), self.pytree_static
        )


class NonArrayStaticTransform(AbstractPyTreeTransform[T]):
    """Wraps non-arrays into the PyTree structure."""

    pytree_dynamic: T
    pytree_static: T = field(static=True)

    def __init__(self, pytree: T):
        self.pytree_dynamic, self.pytree_static = eqx.partition(pytree, eqx.is_array)

    @property
    @override
    def value(self) -> T:
        return eqx.combine(self.pytree_dynamic, self.pytree_static)



"""
Functions for creating coordinate systems.
"""

import jax.numpy as jnp
import numpy as np
from jaxtyping import Array, Float


def make_coordinate_grid(
    shape: tuple[int, ...],
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""] = 1.0,
) -> Float[Array, "*shape ndim"]:
    """
    Create a real-space cartesian coordinate system on a grid.

    **Arguments:**

    - `shape`:
        Shape of the grid, with `ndim = len(shape)`.
    - `grid_spacing`:
        The grid spacing (i.e. pixel/voxel size),
        in units of length.

    **Returns:**

    A cartesian coordinate system in real space.
    """
    coordinate_grid = _make_coordinates_or_frequencies(
        shape, grid_spacing=grid_spacing, outputs_real_space=True
    )
    return coordinate_grid


def make_radial_coordinate_grid(
    shape: tuple[int, ...],
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""] = 1.0,
) -> Float[Array, " *shape"]:
    """Create a real-space radial coordinate system on a grid.

    This wraps the function `make_coordinate_grid` to compute
    the coordinate vector magnitude.

    **Arguments:**

    - `shape`:
        Shape of the grid, with `ndim = len(shape)`.
    - `grid_spacing`:
        The grid spacing (i.e. pixel/voxel size),
        in units of length.

    **Returns:**

    A radial coordinate system in real space.
    """
    # Make a cartesian grid
    coordinate_grid = make_coordinate_grid(
        shape=shape,
        grid_spacing=grid_spacing,
    )
    # Now compute the magnitude of the coordinate vector
    radial_coordinate_grid = jnp.linalg.norm(coordinate_grid, axis=-1)

    return radial_coordinate_grid


def make_frequency_grid(
    shape: tuple[int, ...],
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""] = 1.0,
    outputs_rfftfreqs: bool = True,
) -> Float[Array, "*shape ndim"]:
    """Create a fourier-space cartesian coordinate system on a grid.
    The zero-frequency component is in the corner.

    **Arguments:**

    - `shape`:
        Shape of the grid, with `ndim = len(shape)`.
    - `grid_spacing`:
        The grid spacing (i.e. pixel/voxel size),
        in units of length.
    - `outputs_rfftfreqs`:
        Return a frequency grid for use with `jax.numpy.fft.rfftn`.
        `shape[-1]` is the axis on which the negative
        frequencies are omitted.

    **Returns:**

    A cartesian coordinate system in frequency space.
    """
    frequency_grid = _make_coordinates_or_frequencies(
        shape,
        grid_spacing=grid_spacing,
        outputs_real_space=False,
        outputs_rfftfreqs=outputs_rfftfreqs,
    )
    return frequency_grid


def make_radial_frequency_grid(
    shape: tuple[int, ...],
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""] = 1.0,
    outputs_rfftfreqs: bool = True,
) -> Float[Array, " *shape"]:
    """Create a fourier-space radial coordinate system on a grid.
    The zero-frequency component is in the corner.

    This wraps the function `make_frequency_grid` to compute
    the frequency magnitude, which is a common use case for
    things like computing fourier shell correlations and power spectrums.

    **Arguments:**

    - `shape`:
        Shape of the grid, with `ndim = len(shape)`.
    - `grid_spacing`:
        The grid spacing (i.e. pixel/voxel size),
        in units of length.
    - `outputs_rfftfreqs`:
        Return a frequency grid for use with `jax.numpy.fft.rfftn`.
        `shape[-1]` is the axis on which the negative
        frequencies are omitted.

    **Returns:**

    A radial coordinate system in frequency space.
    """
    # Make a cartesian grid
    frequency_grid = make_frequency_grid(
        shape=shape, grid_spacing=grid_spacing, outputs_rfftfreqs=outputs_rfftfreqs
    )

    # Now compute the magnitude of the frequency vector
    radial_frequency_grid = jnp.linalg.norm(frequency_grid, axis=-1)

    return radial_frequency_grid


def make_frequency_slice(
    shape: tuple[int, int],
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""] = 1.0,
    outputs_rfftfreqs: bool = False,
) -> Float[Array, "1 {shape[0]} {shape[1]} 3"]:
    """Create a fourier-space cartesian coordinate system on a grid, where
    zero-frequency component is in the *center* of the grid.

    !!! warning
        In the function `make_frequency_grid`, the convention is that
        the grid is returned with the zero frequency component is in the
        corner. In this function, as mentioned above, frequency slices are
        returned with the zero frequency component in the center. To convert
        between the two conventions, run

        ```python
        import jax.numpy as jnp
        from cryojax.coordinates import make_frequency_slice

        frequency_slice_with_zero_in_center = make_frequency_slice((100, 100)) # Shape (1, 100, 100, 3)
        frequency_slice_with_zero_in_corner = jnp.fft.ifftshift(frequency_slice_with_zero_in_center, axes=(1, 2))
        ```

        The reason for the difference is so that this function can be used to
        directly pass a `frequency_slice` to the `cryojax.simulator.FourierVoxelGridPotential`,
        which requires that the zero is in the center of the grid.

    **Arguments:**

    - `shape`:
        Shape of the frequency slice, e.g. `shape = (100, 100)`.
    - `grid_spacing`:
        The grid spacing (i.e. voxel size), in units of length.
    - `outputs_rfftfreqs`:
        Return a frequency grid for use with `jax.numpy.fft.rfftn`.
        `shape[-1]` is the axis on which the negative
        frequencies are omitted.

    **Returns:**

    The central, $q_z = 0$ slice of a 3D frequency grid $(q_x, q_y, q_z)$, where
    zero-frequency component is in the *center* of the grid.
    """  # noqa: E501
    frequency_slice = make_frequency_grid(
        shape, grid_spacing, outputs_rfftfreqs=outputs_rfftfreqs
    )
    if outputs_rfftfreqs:
        frequency_slice = jnp.fft.fftshift(frequency_slice, axes=(0,))
    else:
        frequency_slice = jnp.fft.fftshift(frequency_slice, axes=(0, 1))
    frequency_slice = jnp.expand_dims(
        jnp.pad(
            frequency_slice,
            ((0, 0), (0, 0), (0, 1)),
            mode="constant",
            constant_values=0.0,
        ),
        axis=0,
    )
    return frequency_slice


def make_1d_coordinate_grid(
    size: int,
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""] = 1.0,
) -> Float[Array, "*shape ndim"]:
    """
    Create a 1D real-space cartesian coordinate array.

    **Arguments:**

    - `size`:
        Size of the coordinate array.
    - `grid_spacing`:
        The grid spacing (i.e. pixel/voxel size),
        in units of length.

    **Returns:**

    A 1D cartesian coordinate array in real space.
    """
    coordinate_array = _make_coordinates_or_frequencies_1d(
        size, grid_spacing=grid_spacing, outputs_real_space=True
    )
    return coordinate_array


def make_1d_frequency_grid(
    size: int,
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""] = 1.0,
    outputs_rfftfreqs: bool = True,
) -> Float[Array, "*shape ndim"]:
    """Create a 1D fourier-space cartesian coordinate array.
    If `outputs_rfftfreqs = False`, the zero-frequency component is in the beginning.

    Arguments
    ---------
    - `size`:
        Size of the coordinate array.
    - `grid_spacing`:
        The grid spacing (i.e. pixel/voxel size),
        in units of length.
    - `outputs_rfftfreqs`:
        Return a frequency grid for use with `jax.numpy.fft.rfftn`.
        `shape[-1]` is the axis on which the negative
        frequencies are omitted.

    **Returns:**

    A 1D cartesian coordinate array in frequency space.
    """
    frequency_array = _make_coordinates_or_frequencies_1d(
        size,
        grid_spacing=grid_spacing,
        outputs_real_space=False,
        outputs_rfftfreqs=outputs_rfftfreqs,
    )
    return frequency_array


def _make_coordinates_or_frequencies(
    shape: tuple[int, ...],
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""] = 1.0,
    outputs_real_space: bool = False,
    outputs_rfftfreqs: bool = True,
) -> Float[Array, "*shape ndim"]:
    ndim = len(shape)
    coords1D = []
    for idx in range(ndim):
        if outputs_real_space:
            c1D = _make_coordinates_or_frequencies_1d(
                shape[idx], grid_spacing, outputs_real_space
            )
        else:
            if not outputs_rfftfreqs:
                rfftfreq = False
            else:
                rfftfreq = False if idx < ndim - 1 else True
            c1D = _make_coordinates_or_frequencies_1d(
                shape[idx], grid_spacing, outputs_real_space, rfftfreq
            )
        coords1D.append(c1D)
    if ndim == 2:
        y, x = coords1D
        xv, yv = jnp.meshgrid(x, y, indexing="xy")
        coords = jnp.stack([xv, yv], axis=-1)
    elif ndim == 3:
        z, y, x = coords1D
        xv, yv, zv = jnp.meshgrid(x, y, z, indexing="xy")
        xv, yv, zv = [
            jnp.transpose(rv, axes=[2, 0, 1]) for rv in [xv, yv, zv]
        ]  # Change axis ordering to [z, y, x]
        coords = jnp.stack([xv, yv, zv], axis=-1)
    else:
        raise ValueError(
            "Only 2D and 3D coordinate grids are supported. "
            f"Tried to create a grid of shape {shape}."
        )

    return coords


def _make_coordinates_or_frequencies_1d(
    size: int,
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""],
    outputs_real_space: bool = False,
    outputs_rfftfreqs: bool | None = None,
) -> Float[Array, " size"]:
    """One-dimensional coordinates in real or fourier space"""
    if outputs_real_space:
        make_1d = lambda size, dx: jnp.fft.fftshift(jnp.fft.fftfreq(size, 1 / dx)) * size
    else:
        if outputs_rfftfreqs is None:
            raise ValueError("Internal error in `cryojax.coordinates`.")
        else:
            fn = jnp.fft.rfftfreq if outputs_rfftfreqs else jnp.fft.fftfreq
            make_1d = lambda size, dx: fn(size, dx)

    return make_1d(size, grid_spacing)



"""Functionality in `cryojax` for representing datasets."""

import abc
from typing import Generic, TypeVar


T = TypeVar("T")


class AbstractDataset(abc.ABC, Generic[T]):
    """An abstraction of a dataset in `cryojax`. To create an
    `AbstractDataset`, implement its `__init__`, `__getitem__`, and
    `__len__` methods.

    This follows the
    [`torch.utils.data.Dataset`](https://pytorch.org/docs/stable/data.html#torch.utils.data.Dataset)
    API and can easily be wrapped into a pytorch `Dataset` with something like the
    following pattern:

    ```python
    import numpy as np
    from torch.utils.data import Dataset

    class CustomTorchDataset(Dataset):

        def __init__(cryojax_dataset: AbstractDataset):
            self.cryojax_dataset = cryojax_dataset

        def __getitem___(self, index) -> dict[str, np.ndarray]:
            particle_info = self.cryojax_dataset[index]
            return dict(index=index, images=np.asarray(particle_info["images"]))

        def __len__(self) -> int:
            return len(self.cryojax_dataset)
    ```

    JAX also includes packages for dataloaders, such as
    [`jax-dataloaders`](https://github.com/BirkhoffG/jax-dataloader/tree/main) and
    [`grain`](https://github.com/google/grain).

    !!! question "How do I implement an `AbstractDataset`?"

        Implementing an `AbstractDataset` is not like implementing
        most other classes in `cryojax`, which are `equinox.Module`s.
        An `equinox.Module` is just a pytree, so it can be safely
        passed to `jax` transformations. However, an `AbstractDataset`
        can *not* be passed to `jax` transformations. Therefore, it is
        not a pytree. Rather, it is a normal python class.

    """

    @abc.abstractmethod
    def __getitem__(self, index) -> T:
        raise NotImplementedError

    @abc.abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError



import cryojax.simulator as cxs
import equinox as eqx
import numpy as np
import pytest


@pytest.mark.parametrize(
    "pixel_size, shape, ctf_params",
    (
        (
            1.0,
            (75, 75),
            (0.1, 300.0, 2500.0, -100.0, 10.0),
        ),
        (
            1.0,
            (75, 75),
            (0.1, 300.0, 2500.0, -100.0, 10.0),
        ),
        (
            1.0,
            (75, 75),
            (0.1, 300.0, 2500.0, -100.0, 10.0),
        ),
    ),
)
def test_scattering_theories_no_pose(
    sample_pdb_path,
    pixel_size,
    shape,
    ctf_params,
):
    (
        ac,
        voltage_in_kilovolts,
        defocus_in_angstroms,
        astigmatism_in_angstroms,
        astigmatism_angle,
    ) = ctf_params

    atom_potential = cxs.load_tabulated_volume(
        sample_pdb_path,
        output_type=cxs.GaussianMixtureVolume,
        selection_string="not element H",
    )
    instrument_config = cxs.BasicImageConfig(
        shape=shape,
        pixel_size=pixel_size,
        voltage_in_kilovolts=voltage_in_kilovolts,
    )
    pose = cxs.EulerAnglePose()

    ctf = cxs.AstigmaticCTF(
        defocus_in_angstroms=defocus_in_angstroms,
        astigmatism_in_angstroms=astigmatism_in_angstroms,
        astigmatism_angle=astigmatism_angle,
    )
    sp = cxs.IntensityImageModel(
        atom_potential,
        pose,
        instrument_config,
        cxs.StrongPhaseScatteringTheory(
            cxs.GaussianMixtureProjection(sampling_mode="average"),
            cxs.WaveTransferTheory(ctf),
            amplitude_contrast_ratio=ac,
        ),
    )
    wp = cxs.IntensityImageModel(
        atom_potential,
        pose,
        instrument_config,
        cxs.WeakPhaseScatteringTheory(
            cxs.GaussianMixtureProjection(sampling_mode="average"),
            cxs.ContrastTransferTheory(ctf, amplitude_contrast_ratio=ac),
        ),
    )
    # TODO: use jax.linearize for exact agreement
    np.testing.assert_allclose(simulate_fn(sp), simulate_fn(wp), atol=1e-2)


@pytest.mark.parametrize(
    "pixel_size, shape, euler_pose_params, ctf_params",
    (
        (
            1.0,
            (75, 75),
            (2.5, -5.0, 0.0, 0.0, 0.0),
            (0.1, 300.0, 2500.0, -100.0, 10.0),
        ),
        (
            1.0,
            (75, 75),
            (0.0, 0.0, 10.0, -30.0, 60.0),
            (0.1, 300.0, 2500.0, -100.0, 10.0),
        ),
        (
            1.0,
            (75, 75),
            (2.5, -5.0, 10.0, -30.0, 60.0),
            (0.1, 300.0, 2500.0, -100.0, 10.0),
        ),
    ),
)
def test_scattering_theories_pose(
    sample_pdb_path,
    pixel_size,
    shape,
    euler_pose_params,
    ctf_params,
):
    (
        ac,
        voltage_in_kilovolts,
        defocus_in_angstroms,
        astigmatism_in_angstroms,
        astigmatism_angle,
    ) = ctf_params

    atom_potential = cxs.load_tabulated_volume(
        sample_pdb_path,
        output_type=cxs.GaussianMixtureVolume,
        selection_string="name CA ",
    )
    instrument_config = cxs.BasicImageConfig(
        shape=shape,
        pixel_size=pixel_size,
        voltage_in_kilovolts=voltage_in_kilovolts,
    )
    pose = cxs.EulerAnglePose(*euler_pose_params)

    ctf = cxs.AstigmaticCTF(
        defocus_in_angstroms=defocus_in_angstroms,
        astigmatism_in_angstroms=astigmatism_in_angstroms,
        astigmatism_angle=astigmatism_angle,
    )

    sp = cxs.IntensityImageModel(
        atom_potential,
        pose,
        instrument_config,
        cxs.StrongPhaseScatteringTheory(
            cxs.GaussianMixtureProjection(sampling_mode="average"),
            cxs.WaveTransferTheory(ctf),
            amplitude_contrast_ratio=ac,
        ),
    )
    wp = cxs.IntensityImageModel(
        atom_potential,
        pose,
        instrument_config,
        cxs.WeakPhaseScatteringTheory(
            cxs.GaussianMixtureProjection(sampling_mode="average"),
            cxs.ContrastTransferTheory(ctf, amplitude_contrast_ratio=ac),
        ),
    )
    # TODO: use jax.linearize for exact agreement
    np.testing.assert_allclose(simulate_fn(sp), simulate_fn(wp), atol=1e-2)


@eqx.filter_jit
def simulate_fn(model):
    return model.simulate()



from ._base_transform import (
    AbstractImageTransform as AbstractImageTransform,
    ScaleImage as ScaleImage,
)
from ._filters import (
    AbstractFilter as AbstractFilter,
    CustomFilter as CustomFilter,
    HighpassFilter as HighpassFilter,
    LowpassFilter as LowpassFilter,
    WhiteningFilter as WhiteningFilter,
)
from ._masks import (
    AbstractMask as AbstractMask,
    CircularCosineMask as CircularCosineMask,
    CustomMask as CustomMask,
    Cylindrical2DCosineMask as Cylindrical2DCosineMask,
    Rectangular2DCosineMask as Rectangular2DCosineMask,
    Rectangular3DCosineMask as Rectangular3DCosineMask,
    SincCorrectionMask as SincCorrectionMask,
    SphericalCosineMask as SphericalCosineMask,
    SquareCosineMask as SquareCosineMask,
)
from ._spatial_transform import (
    PhaseShiftFFT as PhaseShiftFFT,
    RotateFFT as RotateFFT,
)



"""Routines for downsampling arrays using fourier cropping."""

import math
from collections.abc import Callable

import equinox as eqx
import jax.numpy as jnp
from jax import lax
from jaxtyping import Array, Complex, Float, Inexact

from ..jax_util import NDArrayLike
from ._coordinates import make_frequency_grid
from ._edges import crop_to_shape
from ._fft import fftn, ifftn, rfftn


def block_reduce_downsample(
    image_or_volume: Inexact[NDArrayLike, "_ _"] | Inexact[NDArrayLike, "_ _ _"],
    downsample_factor: int,
    operation: Callable[[Array, Array], Array] = lax.add,
    center_correct: bool = True,
) -> Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"]:
    """Downsample an array by pooling together blocks.
    Wraps `equinox.nn.Pool`.

    **Arguments:**

    - `image_or_volume`:
        image or volume array to downsample. The shape must be
        a multiple of `downsample_factor`
    - `downsample_factor`:
        A scale factor at which to downsample `image_or_volume`
        by. Must be a value greater than `1`.
    - `operation`:
        A function such as `operation = lambda x, y: f(x, y)`,
        where `x` and `y` are JAX arrays. See [`equinox.nn.Pool`]
        (https://docs.kidger.site/equinox/api/nn/pool/#equinox.nn.Pool)
        for documentation.
    - `center_correct`:
        If `True`, apply a phase shift in the fourier domain to correct
        the array center after downsampling. Applies only to even
        `downsample_factor`.

    **Returns:**

    The downsampled `image_or_volume` at shape reduced by
    `downsample_factor`.
    """
    array, k = image_or_volume, downsample_factor
    if k < 1:
        raise ValueError(
            "Called `block_reduce_downsample` with `downsample_factor` less than 1."
        )
    if array.ndim not in [2, 3]:
        raise ValueError(
            "`block_reduce_downsample` was passed an array with "
            f"`ndim = {array.ndim}`, but this function "
            "only supports images and volumes as input."
        )
    if any(s % k != 0 for s in array.shape):
        raise ValueError(
            "`block_reduce_downsample` only supports "
            "downsampling arrays with dimensions that "
            "are a multiple of `downsample_factor`."
            f"Got `downsample_factor = {downsample_factor}` "
            f"but `shape = {array.shape}`."
        )
    # Pooling function downsamples array
    shape = array.shape
    target_shape = tuple(s // k for s in shape)
    kernel_size = array.ndim * (k,)
    if k % 2 == 1:
        padding = tuple(
            ((k - 1) // 2, (k - 1) // 2) if s % 2 == 0 else (0, 0)
            for k, s in zip(kernel_size, shape)
        )
    else:
        padding = tuple((0, 0) for _ in shape)
        if center_correct:
            is_complex = jnp.iscomplexobj(array)
            q = make_frequency_grid(array.shape, outputs_rfftfreqs=False)
            if len(set(target_shape)) > 1:
                raise NotImplementedError(
                    "Tried to call `block_reduce_downsample` "
                    "with `center_correct = True`, even `downsample_factor`, "
                    "and a non-square image/volume. This is not implemented."
                )
            dim = target_shape[0]
            if dim % 2 == 0:
                shift = jnp.full((array.ndim,), (k - 1) / 2)
            else:
                shift = jnp.full((array.ndim,), -0.5)
            phase_shift = jnp.exp(-1.0j * (2 * jnp.pi * jnp.matmul(q, shift)))
            array = ifftn(phase_shift * fftn(array))
            if not is_complex:
                array = array.real
    block_reduce_fn = lambda x: eqx.nn.Pool(
        init=jnp.asarray(0.0, array.dtype),
        operation=operation,
        num_spatial_dims=array.ndim,
        kernel_size=kernel_size,
        stride=kernel_size,
        padding=padding,
        use_ceil=False,
    )(x[None, ...])[0]

    array_ds = block_reduce_fn(array)

    return array_ds


def fourier_crop_downsample(
    image_or_volume: Inexact[NDArrayLike, "_ _"] | Inexact[NDArrayLike, "_ _ _"],
    downsample_factor: float | int,
    outputs_real_space: bool = True,
    preserve_mean: bool = False,
) -> Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"]:
    """Downsample an array using fourier cropping.

    **Arguments:**

    - `image_or_volume`: The image or volume array to downsample.
    - `downsample_factor`:
        A scale factor at which to downsample `image_or_volume`
        by. Must be a value greater than `1`.
    - `outputs_real_space`:
        If `False`, the `image_or_volume` is returned in fourier space
        with the zero-frequency component in the corner. For real signals,
        hermitian symmetry is assumed.
    - `preserve_mean`:
        Preserve the mean of the volume after downsampling, rather
        than the sum.

    **Returns:**

    The downsampled `image_or_volume` at shape reduced by
    `downsample_factor`.
    """
    downsample_factor = float(downsample_factor)
    if downsample_factor < 1.0:
        raise ValueError(
            "Called `fourier_crop_downsample` with `downsample_factor` less than 1."
        )
    if image_or_volume.ndim == 2:
        image = image_or_volume
        new_shape = (
            int(image.shape[0] / downsample_factor),
            int(image.shape[1] / downsample_factor),
        )
        downsampled_array = fourier_crop_to_shape(
            image,
            new_shape,
            preserve_mean=preserve_mean,
            outputs_real_space=outputs_real_space,
        )
    elif image_or_volume.ndim == 3:
        volume = image_or_volume
        new_shape = (
            int(volume.shape[0] / downsample_factor),
            int(volume.shape[1] / downsample_factor),
            int(volume.shape[2] / downsample_factor),
        )
        downsampled_array = fourier_crop_to_shape(
            volume,
            new_shape,
            preserve_mean=preserve_mean,
            outputs_real_space=outputs_real_space,
        )
    else:
        raise ValueError(
            "`fourier_crop_downsample` was passed an array with "
            f"`ndim = {image_or_volume.ndim}`, but this function "
            "only supports images and volumes as input."
        )

    return downsampled_array


def fourier_crop_to_shape(
    image_or_volume: Inexact[NDArrayLike, "_ _"] | Inexact[NDArrayLike, "_ _ _"],
    shape: tuple[int, int] | tuple[int, int, int],
    outputs_real_space: bool = True,
    preserve_mean: bool = False,
) -> Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"]:
    """Downsample an array to a specified shape using fourier cropping.

    For real signals, the Hartley Transform is used to downsample the signal.
    For complex signals, the Fourier Transform is used to downsample the signal.

    The real case is based on the `downsample_transform` function in cryoDRGN
    https://github.com/ml-struct-bio/cryodrgn/blob/4ba75502d4dd1d0e5be3ecabf4a005c652edf4b5/cryodrgn/commands/downsample.py#L154

    **Arguments:**

    - `image_or_volume`: The image or volume array to downsample.
    - `shape`:
        The new shape after fourier cropping.
    - `outputs_real_space`:
        If `False`, the `image_or_volume` is returned in fourier space
        with the zero-frequency component in the corner. For real signals,
        hermitian symmetry is assumed.
    - `preserve_mean`:
        Preserve the mean of the volume after downsampling, rather
        than the sum.

    **Returns:**

    The downsampled `image_or_volume`, at the new real-space shape
    `shape`.
    """
    if jnp.iscomplexobj(image_or_volume):
        signal = _fft_ds_complex_signal_to_shape(
            image_or_volume, shape, outputs_real_space=outputs_real_space
        )
    else:
        signal = _fft_ds_real_signal_to_shape(
            image_or_volume, shape, outputs_real_space=outputs_real_space
        )
    n_pixels, n_pixels_ds = math.prod(image_or_volume.shape), math.prod(shape)
    return (n_pixels_ds / n_pixels) * signal if preserve_mean else signal


def _fft_ds_real_signal_to_shape(
    image_or_volume: Float[NDArrayLike, "_ _"] | Float[NDArrayLike, "_ _ _"],
    downsampled_shape: tuple[int, int] | tuple[int, int, int],
    outputs_real_space: bool = True,
) -> Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"]:
    # Forward Hartley Transform
    hartley_array = jnp.fft.fftshift(fftn(image_or_volume))
    hartley_array = hartley_array.real - hartley_array.imag

    # Crop to the desired shape
    ds_array = crop_to_shape(hartley_array, downsampled_shape)

    # Inverse Hartley Transform
    ds_array = jnp.fft.fftshift(fftn(ds_array))
    ds_array /= ds_array.size
    ds_array = ds_array.real - ds_array.imag

    if outputs_real_space:
        return ds_array
    else:
        return rfftn(ds_array)


def _fft_ds_complex_signal_to_shape(
    image_or_volume: Complex[NDArrayLike, "_ _"] | Complex[NDArrayLike, "_ _ _"],
    downsampled_shape: tuple[int, int] | tuple[int, int, int],
    outputs_real_space: bool = True,
) -> Complex[Array, "_ _"] | Complex[Array, "_ _ _"]:
    fourier_array = jnp.fft.fftshift(fftn(image_or_volume))

    # Crop to the desired shape
    cropped_fourier_array = crop_to_shape(fourier_array, downsampled_shape)

    if outputs_real_space:
        return ifftn(jnp.fft.ifftshift(cropped_fourier_array))
    else:
        return jnp.fft.ifftshift(cropped_fourier_array)



from collections.abc import Callable
from typing import TypeVar

import equinox as eqx
import jax
from jaxtyping import Array, PyTree, Shaped

from ._pytree_transforms import NonArrayStaticTransform


X = TypeVar("X")
Y = TypeVar("Y")
Carry = TypeVar("Carry")


def filter_bmap(
    f: Callable[
        [PyTree[Shaped[Array, "_ ..."], "X"]], PyTree[Shaped[Array, "_ ..."], "Y"]
    ],
    xs: PyTree[Shaped[Array, "_ ..."], "X"],
    *,
    batch_size: int = 1,
) -> PyTree[Shaped[Array, "_ ..."], "Y"]:
    """Like `jax.lax.map(..., batch_size=...)`, but accepts `x`
    with the same rank as `xs`. `xs` is filtered in the usual
    `equinox.filter_*` way.

    **Arguments:**

    - `f`:
        As `jax.lax.map` with format `f(x)`, except
        vmapped over the first axis of the arrays of `x`.
    - `xs`:
        As `jax.lax.map`.
    - `batch_size`:
        Compute a loop of vmaps over `xs` in chunks of `batch_size`.

    **Returns:**

    As `jax.lax.map`.
    """

    def f_scan(carry, x):
        return carry, f(x)

    _, ys = filter_bscan(f_scan, None, xs, batch_size=batch_size)

    return ys


def filter_bscan(
    f: Callable[[Carry, X], tuple[Carry, Y]],
    init: Carry,
    xs: X,
    length: int | None = None,
    unroll: int | bool = 1,
    *,
    batch_size: int = 1,
) -> tuple[Carry, Y]:
    """Like `jax.lax.map(..., batch_size=...)`, except adding
    a `batch_size` to `jax.lax.scan`. Additionally, unlike
    `jax.lax.map`, `f(carry, x)` accepts `x` with the same
    rank as `xs` (e.g. perhaps it is vmapped over `x`).
    `xs` and `carry` are filtered in the usual `equinox.filter_*` way.

    **Arguments:**

    - `f`:
        As `jax.lax.scan` with format `f(carry, x)`.
    - `init`:
        As `jax.lax.scan`.
    - `xs`:
        As `jax.lax.scan`.
    - `length`:
        As `jax.lax.scan`.
    - `unroll`:
        As `jax.lax.scan`.
    - `batch_size`:
        Compute a loop of vmaps over `xs` in chunks of `batch_size`.

    **Returns:**

    As `jax.lax.scan`.
    """
    tree_leaves = jax.tree.leaves(eqx.filter(xs, eqx.is_array))
    if len(tree_leaves) == 0:
        raise ValueError(
            "Called `cryojax.jax_util.filter_bscan` with `xs` "
            "containing no JAX/numpy arrays. Unlike regular `jax.lax.scan` "
            "`xs` is not optional."
        )
    if any(leaf.shape == () for leaf in tree_leaves):
        raise ValueError(
            "Called `cryojax.jax_util.filter_bscan` with `xs` "
            "containing JAX/numpy array scalars (i.e. `shape = ()`). "
            "All JAX/numpy arrays in `xs` must have "
            "a leading dimension that are equal to one another."
        )
    batch_dim = tree_leaves[0].shape[0]
    if not all(leaf.shape[0] == batch_dim for leaf in tree_leaves):
        raise ValueError(
            "Called `cryojax.jax_util.filter_bscan` with `xs` "
            "containing JAX/numpy arrays with different leading "
            "dimensions. All JAX/numpy arrays in `xs` must have "
            "the same leading dimension."
        )
    n_batches = batch_dim // batch_size
    # Filter
    xs_dynamic, xs_static = eqx.partition(xs, eqx.is_array)
    init_dynamic, init_static = eqx.partition(init, eqx.is_array)

    def f_scan(_carry_dynamic, _xs_dynamic):
        _carry, _xs = (
            eqx.combine(_carry_dynamic, init_static),
            eqx.combine(_xs_dynamic, xs_static),
        )
        _carry, _ys = f(_carry, _xs)
        _carry_dynamic = eqx.filter(_carry, eqx.is_array)
        _ys_wrapped = NonArrayStaticTransform(_ys)
        return _carry_dynamic, _ys_wrapped

    # Scan over batches
    xs_reshaped = jax.tree.map(
        lambda x: x[: batch_dim - batch_dim % batch_size, ...].reshape(
            (n_batches, batch_size, *x.shape[1:])
        ),
        xs_dynamic,
    )
    carry_dynamic, ys_wrapped = jax.lax.scan(
        f_scan, init_dynamic, xs_reshaped, length=length, unroll=unroll
    )
    ys_wrapped = jax.tree.map(
        lambda y: y.reshape(n_batches * batch_size, *y.shape[2:]), ys_wrapped
    )
    if batch_dim % batch_size != 0:
        xs_remainder = jax.tree.map(
            lambda x: x[batch_dim - batch_dim % batch_size :, ...], xs_dynamic
        )
        carry_dynamic, ys_remainder = f_scan(carry_dynamic, xs_remainder)
        ys_wrapped = jax.tree.map(
            lambda x, y: jax.lax.concatenate([x, y], dimension=0),
            ys_wrapped,
            ys_remainder,
        )

    return eqx.combine(carry_dynamic, init_static), ys_wrapped.value
